from functools import lru_cache

from ..datatypes import AudioT


def validate_sox_effects(effects: list[str]) -> None:
    import subprocess

    import numpy as np

    zeros = np.zeros((1, 100), dtype=np.float32)

    for effect in effects:
        name = effect.split()[0]
        if name not in list_sox_effects():
            raise ValueError(f"Effect {name} is not supported.")

    args_list = _build_sox_args(effects)

    for args in args_list:
        # print(f"Validating sox effects: {' '.join(args)}")

        process_handle = subprocess.Popen(  # noqa: S603
            args,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        _, stderr = process_handle.communicate(zeros.T.tobytes(order="F"))
        stderr = stderr.decode("utf-8")
        status = process_handle.returncode

        if status != 0:
            raise ValueError(f"For sox effects: {' '.join(effects)}\n{stderr}")


def apply_sox_effects(audio: AudioT, effects: list[str]) -> AudioT:
    """Apply effects to audio data using sox

    :param audio: Audio
    :param effects: List of effects
    :return: Effected audio
    """
    import subprocess

    import numpy as np

    new_audio = audio.copy()

    args_list = _build_sox_args(effects)
    for args in args_list:
        # print(f"Applying sox effects: {' '.join(args)}")

        process_handle = subprocess.Popen(  # noqa: S603
            args,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        stdout, stderr = process_handle.communicate(new_audio.T.tobytes(order="F"))
        stderr = stderr.decode("utf-8")
        status = process_handle.returncode

        if status != 0:
            raise RuntimeError(stderr)

        old_samples = len(new_audio)
        new_audio = np.frombuffer(stdout, dtype=audio.dtype)

        # The length sometimes changes +/-1 with the 'pitch' effect;
        # force the output back to the original length
        new_samples = len(new_audio)
        if "pitch" in args:
            if abs(new_samples - old_samples) > 1:
                raise RuntimeError(
                    "Encountered unexpected length change during 'pitch' effect:\n"
                    + f"{' '.join(args)}\n"
                    + f"original length: {old_samples}, new length: {new_samples}"
                )
            if new_samples < old_samples:
                new_audio = np.pad(array=new_audio, pad_width=(0, old_samples - new_samples))
            elif new_samples > old_samples:
                new_audio = new_audio[:old_samples]

    return new_audio


def _build_sox_args(effects: list[str]) -> list[list[str]]:
    from shlex import split

    from ..constants import SAMPLE_RATE

    base_args = [
        "sox",
        "-D",  # don't dither automatically
        "-V2",  # set verbosity to warning
        "-t",  # set input file type
        "f32",
        "-r",  # set input sample rate
        SAMPLE_RATE,
        "-c",  # set input channels
        1,
        "-",
        "-t",  # set output file type
        "raw",
        "-r",  # set output sample rate
        SAMPLE_RATE,
        "-b",  # set output encoded sample size in bits
        32,
        "-c",  # set output channels
        1,
        "-",
    ]

    result: list[list[str]] = []
    args: list = []
    for effect in effects:
        # If this is a pitch effect and there were other effects already,
        # isolate those other effects and start a new chain
        if effect.startswith("pitch") and args:
            result.append([str(x) for x in base_args + args])
            args = []

        args.extend(split(effect))

        # If this is a pitch effect, finish isolating it as its own chain
        # This allows "fixing" the length after applying the effect
        if effect.startswith("pitch"):
            result.append([str(x) for x in base_args + args])
            args = []

    if args:
        result.append([str(x) for x in base_args + args])

    return result


@lru_cache
def list_sox_effects() -> list[str]:
    from inspect import getmembers
    from inspect import isfunction

    from . import sox_help

    return [member[0] for member in getmembers(sox_help, isfunction)]


def help_sox_effects(name: str) -> str:
    from . import sox_help

    if name not in list_sox_effects():
        raise ValueError(f"Effect {name} not supported.")

    return getattr(sox_help, name)()


def sox_stats(audio: AudioT, win_len: float | None = None) -> str:
    import subprocess

    from ..constants import SAMPLE_RATE

    args = [
        "sox",
        "-D",
        "-V2",
        "-t",
        "f32",
        "-r",
        SAMPLE_RATE,
        "-c",
        1,
        "-",
        "-n",
        "stats",
    ]

    if win_len is not None:
        args.extend(["-w", win_len])

    args = [str(x) for x in args]

    process_handle = subprocess.Popen(  # noqa: S603
        args,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    _, stderr = process_handle.communicate(audio.T.tobytes(order="F"))
    stderr = stderr.decode("utf-8")
    status = process_handle.returncode

    if status != 0:
        raise RuntimeError(stderr)

    return stderr



# ruff: noqa: F821
import numpy as np
import pandas as pd

from ..datatypes import GeneralizedIDs
from ..datatypes import Predict
from ..datatypes import Truth
from ..mixture.mixdb import MixtureDatabase


def class_summary(
    mixdb: MixtureDatabase,
    mixids: GeneralizedIDs,
    truth_f: Truth,
    predict: Predict,
    predict_thr: float | np.ndarray = 0,
    truth_thr: float = 0.5,
    timesteps: int = 0,
) -> pd.DataFrame:
    """Calculate table of metrics per class, and averages for a list
    of mixtures using truth and prediction data [features, num_classes]
    Example:
    Generate multi-class metric summary into table, for example:
                    PPV     TPR      F1     FPR     ACC   AP  AUC  Support
        Class 1     0.71    0.80    0.75    0.00    0.99            44
        Class 2     0.90    0.76    0.82    0.00    0.99            128
        Class 3     0.86    0.82    0.84    0.04    0.93            789
        Other       0.94    0.96    0.95    0.18    0.92            2807

      micro-avg                     0.92    0.027                   3768
      macro avg     0.85    0.83    0.84    0.05    0.96            3768
      micro-avgwo
    """
    from ..metrics.one_hot import one_hot

    num_classes = truth_f.shape[1]

    # TODO: re-work for modern mixdb API
    y_truth_f, y_predict = get_mixids_data(mixdb, mixids, truth_f, predict)  # type: ignore[name-defined]

    if num_classes > 1:
        if not isinstance(predict_thr, np.ndarray):
            if predict_thr == 0:
                predict_thr = np.atleast_1d(0.5)
            else:
                predict_thr = np.atleast_1d(predict_thr)
        else:
            if predict_thr.ndim == 1 and predict_thr[0] == 0:
                predict_thr = np.atleast_1d(0.5)

    _, metrics, _, _, _, metavg = one_hot(y_truth_f, y_predict, predict_thr, truth_thr, timesteps)

    # [ACC, TPR, PPV, TNR, FPR, HITFA, F1, MCC, NT, PT, TP, FP, AP, AUC]
    table_idx = np.array([2, 1, 6, 4, 0, 12, 13, 9])
    col_n = ["PPV", "TPR", "F1", "FPR", "ACC", "AP", "AUC", "Support"]
    if len(mixdb.class_labels) == num_classes:
        row_n = mixdb.class_labels
    else:
        row_n = [f"Class {i}" for i in range(1, num_classes + 1)]

    df = pd.DataFrame(metrics[:, table_idx], columns=col_n, index=row_n)  # pyright: ignore [reportArgumentType]

    # [miPPV, miTPR, miF1, miFPR, miACC, miAP, miAUC, TPSUM]
    avg_row_n = ["Macro-avg", "Micro-avg", "Weighted-avg"]
    dfavg = pd.DataFrame(metavg, columns=col_n, index=avg_row_n)  # pyright: ignore [reportArgumentType]

    # dfblank = pd.DataFrame([''])
    # pd.concat([df, dfblank, dfblank, dfavg])

    classdf = pd.concat([df, dfavg])
    # classdf = classdf.round(2)
    classdf["Support"] = classdf["Support"].astype(int)

    return classdf



"""sonusai audiofe

usage: audiofe [-hvdsp] [--version] [-i INPUT] [-l LENGTH] [-a ASR] [-n NOISEDB]
                        [-w WMODEL] [-o FEATURE] MODEL

options:
    -h, --help
    -v, --verbose                   Be verbose.
    -d, --debug                     Write debug data to H5 file.
    -s, --show                      Display a list of available audio inputs.
    -i INPUT, --input INPUT         Audio source from ALSA or .wav file. See -s or arecord -L. [default: default]
    -l LENGTH, --length LENGTH      Length of audio in seconds. [default: -1].
    -m MODEL, --model MODEL         SonusAI ONNX model applied to the captured audio.
    -n NOISEDB, --noiseadd NOISEDB  Amount of noise to keep in clean audio output. [default: -30]
    -p, --playback                  Enable playback of noisy audio, then the model prediction output audio
    -a ASR, --asr ASR               ASR method to use.
    -w WMODEL, --whisper WMODEL     Model used in whisper, aixplain_whisper and faster_whisper methods. [default: tiny].
    -o FEATURE, --feature-overlap   Run SonusAI model in overlap-streaming mode using FEATURE which is an 8-10 character
                                    string specifying a stride-overlap feature of the same type as the model, i.e. a
                                    model with default feature of hun00ns1 could use hun00nv80 or hun00nv128, etc.

Aaware SonusAI Audio Front End.

Capture LENGTH seconds of audio from INPUT. If LENGTH is < 0, then capture until key is pressed. If INPUT is a valid
audio file name, then use the audio data from the specified file. In this case, if LENGTH is < 0, process entire file;
otherwise, process min(length(INPUT), LENGTH) seconds of audio from INPUT. Audio is saved to
audiofe_capture_<TIMESTAMP>.wav.

If a model is specified, run prediction on audio data from this model. Then compute the inverse transform of the
prediction result and save to audiofe_predict_<TIMESTAMP>.wav.

Also, if a model is specified, save plots of the capture data (time-domain signal and feature) to
audiofe_capture_<TIMESTAMP>.png and predict data (time-domain signal and feature) to
audiofe_predict_<TIMESTAMP>.png.

If an ASR is specified, run ASR on the captured audio and print the results. In addition, if a model was also specified,
run ASR on the predict audio and print the results.  Examples: faster_whisper, google,

If the debug option is enabled, write capture audio, feature, reconstruct audio, predict, and predict audio to
audiofe_<TIMESTAMP>.h5.

"""

import numpy as np

from sonusai.datatypes import AudioT


def main() -> None:
    from docopt import docopt

    from sonusai import __version__ as sai_version
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sai_version, options_first=True)

    verbose = args["--verbose"]
    length = float(args["--length"])
    input_name = args["--input"]
    feature_ovr = args["--feature-overlap"]
    asr_name = args["--asr"]
    whisper_name = args["--whisper"]
    debug = args["--debug"]
    show = args["--show"]
    playback = args["--playback"]
    noiseadd = args["--noiseadd"]
    model_name = args["MODEL"]

    import pyaudio

    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import logger
    from sonusai import update_console_handler
    from sonusai.utils import create_timestamp
    from sonusai.utils import get_input_devices
    from sonusai.utils import load_ort_session

    # Setup logging file
    create_file_handler("audiofe.log")
    update_console_handler(verbose)
    initial_log_messages("audiofe")

    if show:
        logger.info("List of available audio inputs:")
        logger.info("")
        p = pyaudio.PyAudio()
        for name in get_input_devices(p):
            logger.info(f"{name}")
        logger.info("")
        p.terminate()
        # return

    ts = create_timestamp()
    capture_name = f"{ts}-noisy"
    capture_wav = capture_name + ".wav"
    capture_png = capture_name + ".png"
    predict_name = f"{ts}-pred"
    predict_wav = predict_name + ".wav"
    predict_png = predict_name + ".png"
    h5_name = f"{ts}-audiofe.h5"

    if model_name is not None:
        session, options, model_root, hparams, sess_inputs, sess_outputs = load_ort_session(model_name)
        if hparams is None:
            logger.error("Error: ONNX model does not have required SonusAI hyperparameters, cannot proceed.")
            raise SystemExit(1)
        feature_mode = hparams["feature"]
        if feature_ovr is not None:
            # TBD checks for match and valid feature_ovr
            stride = int(feature_ovr[7:])
            sov_type = feature_ovr[6]  # v,e,f,t supported, need to calculate stride from tstep
            if sov_type == "v":
                feat_step = int(np.ceil(0.5 * stride))
            elif sov_type == "e":
                feat_step = int(np.ceil(4 * stride / 5))
            elif sov_type == "f":
                feat_step = int(np.ceil(3 * stride / 4))
            elif sov_type == "t":
                feat_step = int(np.ceil(2 * stride / 3))
            else:
                logger.error("Override feature does not have a supported overlap mode, exiting.")
                raise SystemExit(1)
            feature_orig = feature_mode
            feature_mode = feature_ovr
            logger.info(
                f"Overriding feature with {feature_ovr} (was {feature_orig}), with stride={stride}, step={feat_step}."
            )
        else:
            feat_step = 1

        from pyaaware import FeatureGenerator

        fg = FeatureGenerator(feature_mode=feature_mode)
        ftn = fg.ftransform_length  # feature transform length
        ftr = fg.ftransform_overlap  # forward transform samples per step (R)
        fstride = fg.stride  # feature stride
        fsamples = fstride * ftr  # total samples in feature

        in0name = sess_inputs[0].name
        in0type = sess_inputs[0].type
        out_names = [n.name for n in session.get_outputs()]
        if len(sess_inputs) != 1:
            logger.error(f"Error: ONNX model does not have 1 input, but {len(sess_inputs)}. Exit due to unknown input.")
            raise SystemExit(1)
        if verbose:
            logger.info(f"Read and compiled ONNX model from {model_name}.")
            import onnx

            omodel = onnx.load(model_name)
            from sonusai.utils.onnx_utils import get_and_check_inputs
            from sonusai.utils.onnx_utils import get_and_check_outputs

            logger.info(f"Onnx model uses ir_version {omodel.ir_version}")
            onnx_inputs, inshapes = get_and_check_inputs(omodel)  # Note: logs warning if # inputs > 1
            logger.info(f"Onnx model input has {len(inshapes[0])} dims with shape (0 means dynamic): {inshapes[0]}")
            logger.info(f"Onnx model input has type: {in0type}")
            onnx_outputs, oshapes = get_and_check_outputs(omodel)
            logger.info(f"Onnx model output has {len(oshapes[0])} dims with shape (0 means dynamic): {oshapes[0]}")
            import onnxruntime as ort

            providers = ort.get_available_providers()
            logger.info(f"ONNX runtime available providers: {providers}.")
    else:
        logger.error("No ONNX model provided, exiting.")
        raise SystemExit(1)

    from os.path import exists

    import h5py

    from sonusai.constants import SAMPLE_RATE
    from sonusai.mixture import get_audio_from_feature
    from sonusai.mixture import get_feature_from_audio
    from sonusai.utils import calc_asr
    from sonusai.utils import write_audio

    if input_name is not None and exists(input_name):
        capture_audio = get_frames_from_file(input_name, length)
    else:
        try:
            capture_audio = get_frames_from_device(input_name, length)
        except ValueError as e:
            logger.exception(e)
            return
        # Only write if capture from device, not for file input
        write_audio(capture_wav, capture_audio, SAMPLE_RATE)
        logger.debug("")
        logger.debug(f"Wrote capture audio with shape {capture_audio.shape} to {capture_wav}")

    # Pad audio to transform step size
    padlen_tf = int(np.ceil(len(capture_audio) / ftr)) * ftr - len(capture_audio)
    capture_audio = np.pad(capture_audio, (0, padlen_tf), "constant", constant_values=(0, 0))

    if debug:
        with h5py.File(h5_name, "a") as f:
            if "capture_audio" in f:
                del f["capture_audio"]
            f.create_dataset("capture_audio", data=capture_audio)
        logger.info(f"Wrote capture feature data with shape {capture_audio.shape} to {h5_name}")

    if asr_name is not None:
        logger.info(f"Running ASR on captured audio with {asr_name} ...")
        capture_asr = calc_asr(capture_audio, engine=asr_name, model=whisper_name).text
        logger.info(f"Noisy audio ASR: {capture_asr}")

    if model_name is not None:
        # Pad audio to fill total feature stride * transform stride samples
        padlen = int(np.ceil(len(capture_audio) / fsamples)) * fsamples - len(capture_audio)
        capture_audio_p = np.pad(capture_audio, (0, padlen), "constant", constant_values=(0, 0))

        # feature always frames x stride x feat_params, convert to always Batch x Tsteps x Bins
        feature = get_feature_from_audio(audio=capture_audio_p, feature_mode=feature_mode)
        if debug:
            with h5py.File(h5_name, "a") as f:
                if "feature" in f:
                    del f["feature"]
                f.create_dataset("feature", data=feature)
            logger.info(f"Wrote feature with shape {feature.shape} to {h5_name}")

        feat_nov = sov2nov(feature, feat_step)  # remove overlap, output always Batch x Tsteps x Bins
        # TBD remove padding of feature-stride
        # if padlen > 0:
        save_figure(capture_png, capture_audio, feat_nov)
        logger.info(f"Wrote capture plots to {capture_png}")

        if feature_ovr is not None:
            test_audio = get_audio_from_feature(feature=feat_nov, feature_mode=feature_orig)
            # write_audio(f'{ts}-noisy-itf.wav', test_audio, SAMPLE_RATE)
        else:
            # feature is frames x 1 x Bins, reshape to 1 x frames x Bins for model
            feature = feature.transpose((1, 0, 2))

        if in0type.find("float16") != -1:
            logger.info("Detected input of float16, converting all feature inputs to that type.")
            feature = np.float16(feature)  # type: ignore

        # Run inference, ort session wants batch x timesteps x feat_params, outputs numpy BxTxFP or BxFP
        # Note full reshape not needed here since we assume speech enhancement type model, so a transpose suffices
        logger.info(f"Running model on data with shape {feature.shape} ...")
        if feature_ovr is None:
            predict = session.run(out_names, {in0name: feature})[0]  # standard mode (entire batch)
        else:
            predict = np.zeros(feature.shape)
            for i in range(predict.shape[0]):
                logger.debug(f"running batch: {i}")
                predict[i, :, :] = session.run(out_names, {in0name: feature[i : i + 1, :, :]})[0]

        if debug:
            with h5py.File(h5_name, "a") as f:
                if "predict" in f:
                    del f["predict"]
                f.create_dataset("predict", data=predict)
            logger.info(f"Wrote predict data with shape {predict.shape} to {h5_name}")

        if feature_ovr is not None:
            predict = sov2nov(predict, feat_step)  # always returns batch x tsteps x feat_params
            predict_audio = get_audio_from_feature(feature=predict, feature_mode=feature_orig)
        else:
            predict = predict.transpose((1, 0, 2))  # need transpose to frames x 1 x bins
            predict_audio = get_audio_from_feature(feature=predict, feature_mode=feature_mode)

        if predict_audio.shape[0] > capture_audio.shape[0]:
            predict_audio = predict_audio[0 : (capture_audio.shape[0] - predict_audio.shape[0])]

        if predict_audio.shape[0] < capture_audio.shape[0]:
            capture_audio = capture_audio[0 : (predict_audio.shape[0] - capture_audio.shape[0])]

        if noiseadd is not None:
            ngain = np.power(10, min(float(noiseadd), 0.0) / 20.0)  # limit to gain <1, convert to float
            if ngain < 1.0:  # don't apply if it's 1.0
                logger.info(f"Adding back noise with gain of {ngain} = {noiseadd} db.")
                noise = capture_audio - predict_audio
                predict_audio = predict_audio + ngain * noise

        write_audio(predict_wav, predict_audio, SAMPLE_RATE)
        logger.info(f"Wrote predict audio with shape {predict_audio.shape} to {predict_wav}")
        if debug:
            with h5py.File(h5_name, "a") as f:
                if "predict_audio" in f:
                    del f["predict_audio"]
                f.create_dataset("predict_audio", data=predict_audio)
            logger.info(f"Wrote predict audio with shape {predict_audio.shape} to {h5_name}")

        save_figure(predict_png, predict_audio, predict)
        logger.info(f"Wrote predict plots to {predict_png}")

        if asr_name is not None:
            logger.info(f"Running ASR on model-enhanced audio with {asr_name} ...")
            predict_asr = calc_asr(predict_audio, engine=asr_name, model=whisper_name).text
            logger.info(f"Predict audio ASR: {predict_asr}")

        plot_en = True
        if plot_en is not None:
            import subprocess

            # Construct plot command using spgramd, start the process non-blocking (will leave matplot open)
            command = ["python", "spgramd.py", capture_wav, predict_wav]
            process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)

        if playback is not None:
            import sh

            sh.play(capture_wav)
            sh.play(predict_wav)
            flag_end = False
            while not flag_end:
                choice = input("Press 'r' to replay or 'q' to quit: ").strip().lower()
                if choice == "q":
                    print("Quitting...")
                    flag_end = True
                elif choice == "r":
                    print("Replaying...")
                    sh.play(capture_wav)
                    sh.play(predict_wav)
                else:
                    print("Invalid input. Please try again.")


def get_frames_from_device(input_name: str | None, length: float, chunk: int = 1024) -> AudioT:
    from select import select
    from sys import stdin

    import pyaudio

    from sonusai import logger
    from sonusai.constants import CHANNEL_COUNT
    from sonusai.constants import SAMPLE_RATE
    from sonusai.utils import get_input_device_index_by_name
    from sonusai.utils import get_input_devices

    p = pyaudio.PyAudio()

    input_devices = get_input_devices(p)
    if not input_devices:
        raise ValueError("No input audio devices found")

    if input_name is None:
        input_name = input_devices[0]

    try:
        device_index = get_input_device_index_by_name(p, input_name)
    except ValueError as ex:
        msg = f"Could not find {input_name}\n"
        msg += "Available devices:\n"
        for input_device in input_devices:
            msg += f"  {input_device}\n"
        raise ValueError(msg) from ex

    logger.info(f"Capturing from {p.get_device_info_by_index(device_index).get('name')}")
    stream = p.open(
        format=pyaudio.paFloat32, channels=CHANNEL_COUNT, rate=SAMPLE_RATE, input=True, input_device_index=device_index
    )
    stream.start_stream()

    print()
    print("+---------------------------------+")
    print("| Press Enter to stop             |")
    print("+---------------------------------+")
    print()

    elapsed = 0.0
    seconds_per_chunk = float(chunk) / float(SAMPLE_RATE)
    raw_frames = []
    while elapsed < length or length == -1:
        raw_frames.append(stream.read(num_frames=chunk, exception_on_overflow=False))
        elapsed += seconds_per_chunk
        if select(
            [
                stdin,
            ],
            [],
            [],
            0,
        )[0]:
            stdin.read(1)
            length = elapsed

    stream.stop_stream()
    stream.close()
    p.terminate()
    frames = np.frombuffer(b"".join(raw_frames), dtype=np.float32)
    return frames


def get_frames_from_file(input_name: str, length: float) -> AudioT:
    from sonusai import logger
    from sonusai.constants import SAMPLE_RATE
    from sonusai.mixture import read_audio

    logger.info(f"Capturing from {input_name}")
    frames = read_audio(input_name)
    if length != -1:
        num_frames = int(length * SAMPLE_RATE)
        if len(frames) > num_frames:
            frames = frames[:num_frames]
    return frames


def sov2nov(feature: np.ndarray, step: int) -> np.ndarray:
    """Convert stride-overlap batch x stride x bins to no overlap frames x 1 x bins"""

    stride = feature.shape[1]  # stride, tsteps is set to stride in sov mode
    if stride == 1:
        return feature  # no reshape if stride is already 1
    # else:
    #     hs = feature.shape[1]//2   # half of stride
    #     nb = feature.shape[0]      # batches

    nb = feature.shape[0]
    fout = feature[:, (stride - step) :, :]  # take last
    fout = np.reshape(fout, [step * nb, 1, feature.shape[2]])
    return fout  # np.transpose(fout,[1,0,2])


def save_figure(name: str, audio: np.ndarray, feature: np.ndarray) -> None:
    import matplotlib.pyplot as plt
    from scipy.interpolate import CubicSpline

    from sonusai.constants import SAMPLE_RATE
    from sonusai.utils import unstack_complex

    spectrum = 20 * np.log(np.abs(np.squeeze(unstack_complex(feature)).transpose()) + 1e-7)
    frames = spectrum.shape[1]
    samples = (len(audio) // frames) * frames
    length_in_s = samples / SAMPLE_RATE
    interp = samples // frames

    ts = np.arange(0.0, length_in_s, interp / SAMPLE_RATE)
    t = np.arange(0.0, length_in_s, 1 / SAMPLE_RATE)

    spectrum = CubicSpline(ts, spectrum, axis=-1)(t)

    fig, (ax1, ax2) = plt.subplots(nrows=2)
    ax1.set_title(name)
    ax1.plot(t, audio[:samples])
    ax1.set_ylabel("Signal")
    ax1.set_xlim(0, length_in_s)
    ax1.set_ylim(-1, 1)

    ax2.imshow(spectrum, origin="lower", aspect="auto")
    ax2.set_xticks([])
    ax2.set_ylabel("Feature")

    plt.savefig(name, dpi=300)


if __name__ == "__main__":
    from sonusai import exception_handler
    from sonusai.utils import register_keyboard_interrupt

    register_keyboard_interrupt()
    try:
        main()
    except Exception as e:
        exception_handler(e)


# import subprocess
#
# # Define the arguments
# arg1 = "value1"
# arg2 = "value2"
#
# # Construct the command
# command = ["python", "script.py", arg1, arg2]
#
# # Start the process
# process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
#
# # Optionally, you can communicate with the process later if needed
# # For example, to wait for the process to finish and get the output
# stdout, stderr = process.communicate()
#
# # Check if the process was successful
# if process.returncode == 0:
#     print("Process executed successfully:")
#     print(stdout)
# else:
#     print("Process failed:")
#     print(stderr)



from ...datatypes import Truth
from ..mixdb import MixtureDatabase


def metrics_validate(config: dict) -> None:
    if len(config) == 0:
        raise AttributeError("metrics truth function is missing config")

    parameters = ["metric"]
    for parameter in parameters:
        if parameter not in config:
            raise AttributeError(f"metrics truth function is missing required '{parameter}'")


def metrics_parameters(_feature: str, _num_classes: int, _config: dict) -> int | None:
    return None


def metrics(mixdb: MixtureDatabase, m_id: int, category: str, config: dict) -> Truth:
    """Metadata truth generation function

    Retrieves metrics from target.
    """
    if not isinstance(config["metric"], list):
        m = [config["metric"]]
    else:
        m = config["metric"]
    return mixdb.mixture_metrics(m_id, m)[m[0]][category]



import pyaudio


def get_input_device_index_by_name(p: pyaudio.PyAudio, name: str | None = None) -> int:
    info = p.get_host_api_info_by_index(0)
    device_count = info.get("deviceCount")
    if isinstance(device_count, int):
        for i in range(0, device_count):
            device_info = p.get_device_info_by_host_api_device_index(0, i)
            if name is None:
                device_name = None
            else:
                device_name = device_info.get("name")
            if isinstance(device_name, str) or device_name is None:
                input_channels = device_info.get("maxInputChannels")
                if name == device_name and isinstance(input_channels, int) and input_channels > 0:
                    return i

    raise ValueError(f"Could not find {name}")


def get_input_devices(p: pyaudio.PyAudio) -> list[str]:
    names = []
    info = p.get_host_api_info_by_index(0)
    device_count = info.get("deviceCount")
    if isinstance(device_count, int):
        for i in range(0, device_count):
            device_info = p.get_device_info_by_host_api_device_index(0, i)
            device_name = device_info.get("name")
            if isinstance(device_name, str):
                input_channels = device_info.get("maxInputChannels")
                if isinstance(input_channels, int) and input_channels > 0:
                    names.append(device_name)

    return names


def get_default_input_device(p: pyaudio.PyAudio) -> str:
    info = p.get_host_api_info_by_index(0)
    device_count = info.get("deviceCount")
    if isinstance(device_count, int):
        for i in range(0, device_count):
            device_info = p.get_device_info_by_host_api_device_index(0, i)
            device_name = device_info.get("name")
            if isinstance(device_name, str):
                input_channels = device_info.get("maxInputChannels")
                if isinstance(input_channels, int) and input_channels > 0:
                    return device_name

    raise ValueError("No input audio devices found")



# ruff: noqa: F821
import numpy as np
import pandas as pd

from ..datatypes import GeneralizedIDs
from ..datatypes import Predict
from ..datatypes import Segsnr
from ..datatypes import Truth
from ..mixture.mixdb import MixtureDatabase


def snr_summary(
    mixdb: MixtureDatabase,
    mixid: GeneralizedIDs,
    truth_f: Truth,
    predict: Predict,
    segsnr: Segsnr | None = None,
    predict_thr: float | np.ndarray = 0,
    truth_thr: float = 0.5,
    timesteps: int = 0,
) -> tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, dict]:
    """Calculate average-over-class metrics per SNR over specified mixture list.
    Inputs:
      mixdb        Mixture database
      mixid
      truth_f      Truth/labels [features, num_classes]
      predict      Prediction data / neural net model one-hot out [features, num_classes]
      segsnr       Segmental SNR from SonusAI genft  [transform_frames, 1]
      predict_thr  Decision threshold(s) applied to predict data, allowing predict to be
                   continuous probabilities or decisions
      truth_thr    Decision threshold(s) applied to truth data, allowing truth to be
                   continuous probabilities or decisions
      timesteps

    Default predict_thr=0 will infer 0.5 for multi-label mode (truth_mutex = False), or
    if single-label mode (truth_mutex == True) then ignore and use argmax mode, and
    the confusion matrix is calculated for all classes.

    Returns pandas dataframe (snrdf) of metrics per SNR.
    """
    import warnings

    from ..metrics.one_hot import one_hot
    from ..queries.queries import get_mixids_from_snr

    num_classes = truth_f.shape[1]

    snr_mixids = get_mixids_from_snr(mixdb=mixdb, mixids=mixid)

    # Check predict_thr array or scalar and return final scalar predict_thr value
    if num_classes > 1:
        if not isinstance(predict_thr, np.ndarray):
            if predict_thr == 0:
                # multi-label predict_thr scalar 0 force to 0.5 default
                predict_thr = np.atleast_1d(0.5)
            else:
                predict_thr = np.atleast_1d(predict_thr)
        else:
            if predict_thr.ndim == 1 and len(predict_thr) == 1:
                if predict_thr[0] == 0:
                    # multi-label predict_thr array scalar 0 force to 0.5 default
                    predict_thr = np.atleast_1d(0.5)
                else:
                    # multi-label predict_thr array set to scalar = array[0]
                    predict_thr = predict_thr[0]

    macro_avg = np.zeros((len(snr_mixids), 7), dtype=np.float32)
    micro_avg = np.zeros((len(snr_mixids), 7), dtype=np.float32)
    wghtd_avg = np.zeros((len(snr_mixids), 7), dtype=np.float32)
    ssnr_stats = None
    segsnr_f = None

    if segsnr is not None:
        # prep segsnr if provided, transform frames to feature frames via mean()
        # expected to always be an integer
        feature_frames = int(segsnr.shape[0] / truth_f.shape[0])
        segsnr_f = np.mean(
            np.reshape(segsnr, (truth_f.shape[0], feature_frames)),
            axis=1,
            keepdims=True,
        )
        ssnr_stats = np.zeros((len(snr_mixids), 3), dtype=np.float32)

    for ii, snr in enumerate(snr_mixids):
        # TODO: re-work for modern mixdb API
        y_truth, y_predict = get_mixids_data(mixdb, snr_mixids[snr], truth_f, predict)  # type: ignore[name-defined]
        _, _, _, _, _, mavg = one_hot(y_truth, y_predict, predict_thr, truth_thr, timesteps)

        # mavg macro, micro, weighted: [PPV, TPR, F1, FPR, ACC, mAP, mAUC, TPSUM]
        macro_avg[ii, :] = mavg[0, 0:7]
        micro_avg[ii, :] = mavg[1, 0:7]
        wghtd_avg[ii, :] = mavg[2, 0:7]
        if segsnr is not None:
            # TODO: re-work for modern mixdb API
            y_truth, y_segsnr = get_mixids_data(mixdb, snr_mixids[snr], truth_f, segsnr_f)  # type: ignore[name-defined]
            with warnings.catch_warnings():
                warnings.filterwarnings(action="ignore", message="divide by zero encountered in log10")
                # segmental SNR mean = mixture_snr and target_snr
                ssnr_stats[ii, 0] = 10 * np.log10(np.mean(y_segsnr))  # type: ignore[index]
                # segmental SNR 80% percentile
                ssnr_stats[ii, 1] = 10 * np.log10(np.percentile(y_segsnr, 80, method="midpoint"))  # type: ignore[index]
                # segmental SNR max
                ssnr_stats[ii, 2] = 10 * np.log10(max(y_segsnr))  # type: ignore[index]

    # SNR format: PPV, TPR, F1, FPR, ACC, AP, AUC
    col_n = ["PPV", "TPR", "F1", "FPR", "ACC", "AP", "AUC"]
    snr_macrodf = pd.DataFrame(macro_avg, index=list(snr_mixids.keys()), columns=col_n)  # pyright: ignore [reportArgumentType]
    snr_macrodf.sort_index(ascending=False, inplace=True)

    snr_microdf = pd.DataFrame(micro_avg, index=list(snr_mixids.keys()), columns=col_n)  # pyright: ignore [reportArgumentType]
    snr_microdf.sort_index(ascending=False, inplace=True)

    snr_wghtdf = pd.DataFrame(wghtd_avg, index=list(snr_mixids.keys()), columns=col_n)  # pyright: ignore [reportArgumentType]
    snr_wghtdf.sort_index(ascending=False, inplace=True)

    # Add segmental SNR columns if provided
    if segsnr is not None:
        ssnrdf = pd.DataFrame(
            ssnr_stats,
            index=list(snr_mixids.keys()),  # pyright: ignore [reportArgumentType]
            columns=["SSNRavg", "SSNR80p", "SSNRmax"],  # pyright: ignore [reportArgumentType]
        )
        ssnrdf.sort_index(ascending=False, inplace=True)
        snr_macrodf = pd.concat([snr_macrodf, ssnrdf], axis=1)
        snr_microdf = pd.concat([snr_microdf, ssnrdf], axis=1)
        snr_wghtdf = pd.concat([snr_wghtdf, ssnrdf], axis=1)

    return snr_macrodf, snr_microdf, snr_wghtdf, snr_mixids



"""sonusai vars

usage: vars [-h]

options:
   -h, --help   Display this help.

List custom SonusAI variables.

"""


def main() -> None:
    from docopt import docopt

    from sonusai import __version__ as sai_version
    from sonusai.utils import trim_docstring

    docopt(trim_docstring(__doc__), version=sai_version, options_first=True)

    from os import environ
    from os import getenv

    from sonusai.constants import DEFAULT_NOISE

    print("Custom SonusAI variables:")
    print("")
    print(f"${{default_noise}}: {DEFAULT_NOISE}")
    print("")
    print("SonusAI recognized environment variables:")
    print("")
    print(f"DEEPGRAM_API_KEY {getenv('DEEPGRAM_API_KEY')}")
    print(f"GOOGLE_SPEECH_API_KEY {getenv('GOOGLE_SPEECH_API_KEY')}")
    print("")
    items = ["DEEPGRAM_API_KEY", "GOOGLE_SPEECH_API_KEY"]
    items += [item for item in environ if item.upper().startswith("AIXP_WHISPER_")]


if __name__ == "__main__":
    from sonusai import exception_handler
    from sonusai.utils import register_keyboard_interrupt

    register_keyboard_interrupt()
    try:
        main()
    except Exception as e:
        exception_handler(e)



"""sonusai lsdb

usage: lsdb [-hsa] [-i MIXID] [-c CID] LOC

Options:
    -h, --help
    -i MIXID, --mixid MIXID         Mixture ID(s) to analyze. [default: *].
    -c CID, --class_index CID       Analyze mixtures that contain this class index.
    -s, --sources                   List all source files.
    -a, --all_class_counts          List all class counts.

List mixture data information from a SonusAI mixture database.

Inputs:
    LOC     A SonusAI mixture database directory.

"""

from sonusai.datatypes import GeneralizedIDs
from sonusai.mixture import MixtureDatabase


def lsdb(
    mixdb: MixtureDatabase,
    mixids: GeneralizedIDs = "*",
    class_index: int | None = None,
    list_targets: bool = False,
    all_class_counts: bool = False,
) -> None:
    from sonusai import logger
    from sonusai.constants import SAMPLE_RATE
    from sonusai.queries import get_mixids_from_class_indices
    from sonusai.utils import consolidate_range
    from sonusai.utils import max_text_width
    from sonusai.utils import print_mixture_details
    from sonusai.utils import seconds_to_hms

    desc_len = 24

    total_samples = mixdb.total_samples()
    total_duration = total_samples / SAMPLE_RATE

    logger.info(f"{'Mixtures':{desc_len}} {mixdb.num_mixtures}")
    logger.info(f"{'Duration':{desc_len}} {seconds_to_hms(seconds=total_duration)}")
    logger.info(f"{'Sources':{desc_len}} {mixdb.num_source_files}")
    logger.info(f"{'Feature':{desc_len}} {mixdb.feature}")
    logger.info(
        f"{'Feature shape':{desc_len}} {mixdb.fg_stride} x {mixdb.feature_parameters} "
        f"({mixdb.fg_stride * mixdb.feature_parameters} total params)"
    )
    logger.info(f"{'Feature samples':{desc_len}} {mixdb.feature_samples} samples ({mixdb.feature_ms} ms)")
    logger.info(
        f"{'Feature step samples':{desc_len}} {mixdb.feature_step_samples} samples ({mixdb.feature_step_ms} ms)"
    )
    logger.info(f"{'Feature overlap':{desc_len}} {mixdb.fg_step / mixdb.fg_stride} ({mixdb.feature_step_ms} ms)")
    logger.info(f"{'SNRs':{desc_len}} {mixdb.snrs}")
    logger.info(f"{'Random SNRs':{desc_len}} {mixdb.random_snrs}")
    logger.info(f"{'Classes':{desc_len}} {mixdb.num_classes}")
    # TODO: fix class count
    logger.info(f"{'Class count':{desc_len}} not supported")
    # print_class_count(class_count=class_count, length=desc_len, print_fn=logger.info)
    # TODO: add class weight calculations here
    logger.info("")

    if list_targets:
        logger.info("Source details:")
        for category, sources in mixdb.source_files.items():
            print(f"  {category}:")
            for source in sources:
                logger.info(f"{'    Name':{desc_len}} {source.name}")
                logger.info(f"{'    Truth index':{desc_len}} {source.class_indices}")
            logger.info("")

    if class_index is not None:
        if 0 <= class_index > mixdb.num_classes:
            raise ValueError(f"Given class_index is outside valid range of 1-{mixdb.num_classes}")
        ids = get_mixids_from_class_indices(mixdb=mixdb, predicate=lambda x: x in [class_index])[class_index]
        logger.info(f"Mixtures with class index {class_index}: {ids}")
        logger.info("")

    mixids = mixdb.mixids_to_list(mixids)

    if len(mixids) == 1:
        print_mixture_details(mixdb=mixdb, mixid=mixids[0], desc_len=desc_len, print_fn=logger.info)
        if all_class_counts:
            # TODO: fix class count
            logger.info("All class count not supported")
            # print_class_count(class_count=class_count, length=desc_len, print_fn=logger.info, all_class_counts=True)
    else:
        logger.info(
            f"Calculating statistics from truth_f files for {len(mixids):,} mixtures ({consolidate_range(mixids)})"
        )
        logger.info("Not supported")


def main() -> None:
    from docopt import docopt

    from sonusai import __version__ as sai_version
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sai_version, options_first=True)

    mixid = args["--mixid"]
    class_index = args["--class_index"]
    list_targets = args["--targets"]
    all_class_counts = args["--all_class_counts"]
    location = args["LOC"]

    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import logger
    from sonusai import update_console_handler

    if class_index is not None:
        class_index = int(class_index)

    create_file_handler("lsdb.log")
    update_console_handler(False)
    initial_log_messages("lsdb")

    logger.info(f"Analyzing {location}")

    mixdb = MixtureDatabase(location)
    lsdb(
        mixdb=mixdb,
        mixids=mixid,
        class_index=class_index,
        list_targets=list_targets,
        all_class_counts=all_class_counts,
    )


if __name__ == "__main__":
    from sonusai import exception_handler
    from sonusai.utils import register_keyboard_interrupt

    register_keyboard_interrupt()
    try:
        main()
    except Exception as e:
        exception_handler(e)



from collections.abc import Callable

from ..datatypes import ClassCount
from ..mixture.mixdb import MixtureDatabase


def print_mixture_details(
    mixdb: MixtureDatabase,
    mixid: int | None = None,
    desc_len: int = 1,
    print_fn: Callable = print,
) -> None:
    import numpy as np

    from ..constants import SAMPLE_RATE
    from ..utils.seconds_to_hms import seconds_to_hms

    if mixid is not None:
        if 0 < mixid >= mixdb.num_mixtures:
            raise ValueError(f"Given mixid is outside valid range of 0:{mixdb.num_mixtures - 1}.")

        print_fn(f"Mixture {mixid} details")
        mixture = mixdb.mixture(mixid)
        target_files = [mixdb.target_files[target.file_id] for target in mixture.targets]
        target_augmentations = [target.augmentation for target in mixture.targets]
        noise_file = mixdb.noise_file(mixture.noise.file_id)
        for t_idx, target_file in enumerate(target_files):
            print_fn(f"  Target {t_idx}")
            print_fn(f"{'    Name':{desc_len}} {target_file.name}")
            print_fn(f"{'    Duration':{desc_len}} {seconds_to_hms(target_file.duration)}")
            for truth_name, truth_config in target_file.truth_configs.items():
                print_fn(f"    Truth config: '{truth_name}'")
                print_fn(f"{'      Function':{desc_len}} {truth_config.function}")
                print_fn(f"{'      Stride reduction':{desc_len}} {truth_config.stride_reduction}")
                print_fn(f"{'      Config':{desc_len}} {truth_config.config}")
            print_fn(f"{'    Augmentation':{desc_len}} {target_augmentations[t_idx]}")
        print_fn(f"{'  Samples':{desc_len}} {mixture.samples}")
        print_fn(f"{'  Feature frames':{desc_len}} {mixdb.mixture_feature_frames(mixid)}")
        print_fn(f"{'  Noise file':{desc_len}} {noise_file.name}")
        noise_offset_percent = int(np.round(100 * mixture.noise_offset / float(noise_file.duration * SAMPLE_RATE)))
        print_fn(f"{'  Noise offset':{desc_len}} {mixture.noise_offset} samples ({noise_offset_percent}%)")
        print_fn(f"{'  SNR':{desc_len}} {mixture.snr} dB{' (random)' if mixture.snr.is_random else ''}")
        print_fn(
            f"{'  Target gain':{desc_len}} {[target.gain if not mixture.is_noise_only else 0 for target in mixture.targets]}"
        )
        print_fn(f"{'  Target SNR gain':{desc_len}} {mixture.target_snr_gain}")
        print_fn(f"{'  Noise SNR gain':{desc_len}} {mixture.noise_snr_gain}")
        print_fn("")


def print_class_count(
    class_count: ClassCount,
    length: int,
    print_fn: Callable = print,
    all_class_counts: bool = False,
) -> None:
    from ..utils.max_text_width import max_text_width

    print_fn("Class count:")
    idx_len = max_text_width(len(class_count))
    for idx, count in enumerate(class_count):
        if all_class_counts or count > 0:
            desc = f"  class {idx + 1:{idx_len}}"
            print_fn(f"{desc:{length}} {count}")



from typing import NamedTuple


class WerResult(NamedTuple):
    wer: float
    words: int
    substitutions: float
    deletions: float
    insertions: float


def calc_wer(hypothesis: list[str] | str, reference: list[str] | str) -> WerResult:
    """Computes average word error rate between two texts represented as corresponding strings or lists of strings.

    :param hypothesis: the hypothesis sentence(s) as a string or list of strings
    :param reference: the reference sentence(s) as a string or list of strings
    :return: a WerResult object with error, words, insertions, deletions, substitutions
    """
    import jiwer

    transformation = jiwer.Compose(
        [
            jiwer.ToLowerCase(),
            jiwer.RemovePunctuation(),
            jiwer.RemoveWhiteSpace(replace_by_space=True),
            jiwer.RemoveMultipleSpaces(),
            jiwer.Strip(),
            jiwer.RemoveEmptyStrings(),
            jiwer.ReduceToListOfListOfWords(word_delimiter=" "),
        ]
    )

    if isinstance(reference, str):
        reference = [reference]
    if isinstance(hypothesis, str):
        hypothesis = [hypothesis]

    # jiwer does not allow empty string
    measures = {"insertions": 0, "substitutions": 0, "deletions": 0, "hits": 0}
    if any(len(t) == 0 for t in reference):
        if any(len(t) != 0 for t in hypothesis):
            measures["insertions"] = len(hypothesis)
    else:
        measures = jiwer.compute_measures(
            truth=reference,
            hypothesis=hypothesis,
            truth_transform=transformation,
            hypothesis_transform=transformation,
        )

    errors = measures["substitutions"] + measures["deletions"] + measures["insertions"]
    words = measures["hits"] + measures["substitutions"] + measures["deletions"]

    if words != 0:
        wer = errors / words
        substitutions_rate = measures["substitutions"] / words
        deletions_rate = measures["deletions"] / words
        insertions_rate = measures["insertions"] / words
    else:
        wer = float("inf")
        substitutions_rate = float("inf")
        deletions_rate = float("inf")
        insertions_rate = float("inf")

    return WerResult(
        wer=wer,
        words=int(words),
        substitutions=substitutions_rate,
        deletions=deletions_rate,
        insertions=insertions_rate,
    )



from pyaaware import ForwardTransform

from ..datatypes import AudioF
from ..datatypes import AudioT
from ..datatypes import EnergyF


def compute_energy_f(
    frequency_domain: AudioF | None = None,
    time_domain: AudioT | None = None,
    transform: ForwardTransform | None = None,
) -> EnergyF:
    """Compute the energy in each bin

    Must provide either frequency domain or time domain input. If time domain input is provided, must also provide
    ForwardTransform object to use to convert to frequency domain.

    :param frequency_domain: Frequency domain data [frames, bins]
    :param time_domain: Time domain data [samples]
    :param transform: ForwardTransform object
    :return: Frequency domain per-bin energy data [frames, bins]
    """
    import numpy as np
    import torch

    if frequency_domain is None:
        if time_domain is None:
            raise ValueError("Must provide time or frequency domain input")
        if transform is None:
            raise ValueError("Must provide ForwardTransform object")

        _frequency_domain = transform.execute_all(torch.from_numpy(time_domain))[0].numpy()
    else:
        _frequency_domain = frequency_domain

    frames, bins = _frequency_domain.shape
    result = np.empty((frames, bins), dtype=np.float32)

    for f in range(frames):
        for b in range(bins):
            value = _frequency_domain[f, b]
            result[f, b] = np.real(value) * np.real(value) + np.imag(value) * np.imag(value)

    return result



import numpy as np

from ..datatypes import Predict


def read_predict_data(filename: str) -> Predict:
    """Read predict data from given HDF5 file and return it."""
    import h5py

    from .. import logger

    logger.debug(f"Reading prediction data from {filename}")
    with h5py.File(filename, "r") as f:
        # prediction data is either [frames, num_classes], or [frames, timesteps, num_classes]
        predict = np.array(f["predict"])

        if predict.ndim == 2:
            return predict

        if predict.ndim == 3:
            frames, timesteps, num_classes = predict.shape

            logger.debug(
                f"Reshaping prediction data in {filename} "
                f""
                f"from [{frames}, {timesteps}, {num_classes}] "
                f"to [{frames * timesteps}, {num_classes}]"
            )
            predict = np.reshape(predict, [frames * timesteps, num_classes], order="F")
            return predict

        raise RuntimeError(f"Invalid prediction data dimensions in {filename}")



import os
from pathlib import Path

from .types import TimeAlignedType


def _get_duration(name: str) -> float:
    import soundfile

    try:
        return soundfile.info(name).duration
    except Exception as e:
        raise OSError(f"Error reading {name}: {e}") from e


def load_text(audio: str | os.PathLike[str]) -> TimeAlignedType | None:
    """Load time-aligned text data given a VCTK audio file.

    :param audio: Path to the VCTK audio file.
    :return: A TimeAlignedType object.
    """
    import string

    file = Path(audio).parents[2] / "txt" / Path(audio).parent.name / (Path(audio).stem[:-5] + ".txt")
    if not os.path.exists(file):
        return None

    with open(file, encoding="utf-8") as f:
        line = f.read()

    start = 0
    end = _get_duration(str(audio))
    text = line.strip().lower().translate(str.maketrans("", "", string.punctuation))

    return TimeAlignedType(start, end, text)


def load_speakers(input_dir: Path) -> dict:
    speakers = {}
    with open(input_dir / "speaker-info.txt") as file:
        for line in file:
            if not line.startswith("ID"):
                fields = line.strip().split("(", 1)[0].split()
                speaker_id = fields[0]
                age = fields[1]
                gender = fields[2]
                dialect = " ".join(list(fields[3:]))
                speakers[speaker_id] = {
                    "gender": gender,
                    "dialect": dialect,
                    "age": age,
                }
    return speakers



from functools import lru_cache
from pathlib import Path

from ..datatypes import AudioT


def get_next_noise(audio: AudioT, offset: int, length: int) -> AudioT:
    """Get next sequence of noise data from noise audio

    :param audio: Overall noise audio (entire file's worth of data)
    :param offset: Starting sample
    :param length: Number of samples to get
    :return: Sequence of noise audio data
    """
    import numpy as np

    return np.take(audio, range(offset, offset + length), mode="wrap")


def get_duration(audio: AudioT) -> float:
    """Get duration of audio in seconds

    :param audio: Time domain data [samples]
    :return: Duration of audio in seconds
    """
    from ..constants import SAMPLE_RATE

    return len(audio) / SAMPLE_RATE


def validate_input_file(input_filepath: str | Path) -> None:
    from os.path import exists
    from os.path import splitext

    from soundfile import available_formats

    if not exists(input_filepath):
        raise OSError(f"input_filepath {input_filepath} does not exist.")

    ext = splitext(input_filepath)[1][1:].lower()
    read_formats = [item.lower() for item in available_formats()]
    if ext not in read_formats:
        raise OSError(f"This installation cannot process .{ext} files")


def get_sample_rate(name: str | Path, use_cache: bool = True) -> int:
    """Get sample rate from audio file

    :param name: File name
    :param use_cache: If true, use LRU caching
    :return: Sample rate
    """
    if use_cache:
        return _get_sample_rate(name)
    return _get_sample_rate.__wrapped__(name)


@lru_cache
def _get_sample_rate(name: str | Path) -> int:
    """Get sample rate from audio file using soundfile

    :param name: File name
    :return: Sample rate
    """
    import soundfile
    from pydub import AudioSegment

    from ..utils.tokenized_shell_vars import tokenized_expand

    expanded_name, _ = tokenized_expand(name)

    try:
        if expanded_name.endswith(".mp3"):
            return AudioSegment.from_mp3(expanded_name).frame_rate

        if expanded_name.endswith(".m4a"):
            return AudioSegment.from_file(expanded_name).frame_rate

        return soundfile.info(expanded_name).samplerate
    except Exception as e:
        if name != expanded_name:
            raise OSError(f"Error reading {name} (expanded: {expanded_name}): {e}") from e
        else:
            raise OSError(f"Error reading {name}: {e}") from e


def raw_read_audio(name: str | Path) -> tuple[AudioT, int]:
    import numpy as np
    import soundfile
    from pydub import AudioSegment

    from ..utils.tokenized_shell_vars import tokenized_expand

    expanded_name, _ = tokenized_expand(name)

    try:
        if expanded_name.endswith(".mp3"):
            sound = AudioSegment.from_mp3(expanded_name)
            raw = np.array(sound.get_array_of_samples()).astype(np.float32).reshape((-1, sound.channels))
            raw = raw / 2 ** (sound.sample_width * 8 - 1)
            sample_rate = sound.frame_rate
        elif expanded_name.endswith(".m4a"):
            sound = AudioSegment.from_file(expanded_name)
            raw = np.array(sound.get_array_of_samples()).astype(np.float32).reshape((-1, sound.channels))
            raw = raw / 2 ** (sound.sample_width * 8 - 1)
            sample_rate = sound.frame_rate
        else:
            raw, sample_rate = soundfile.read(expanded_name, always_2d=True, dtype="float32")
    except Exception as e:
        if name != expanded_name:
            raise OSError(f"Error reading {name} (expanded: {expanded_name}): {e}") from e
        else:
            raise OSError(f"Error reading {name}: {e}") from e

    return np.squeeze(raw[:, 0].astype(np.float32)), sample_rate


def read_audio(name: str | Path, use_cache: bool = True) -> AudioT:
    """Read audio data from a file

    :param name: File name
    :param use_cache: If true, use LRU caching
    :return: Array of time domain audio data
    """
    if use_cache:
        return _read_audio(name)
    return _read_audio.__wrapped__(name)


@lru_cache
def _read_audio(name: str | Path) -> AudioT:
    """Read audio data from a file using soundfile

    :param name: File name
    :return: Array of time domain audio data
    """
    from ..constants import SAMPLE_RATE
    from .resample import resample

    out, sample_rate = raw_read_audio(name)

    return resample(out, orig_sr=sample_rate, target_sr=SAMPLE_RATE)


def get_num_samples(name: str | Path, use_cache: bool = True) -> int:
    """Get the number of samples resampled to the SonusAI sample rate in the given file

    :param name: File name
    :param use_cache: If true, use LRU caching
    :return: number of samples in resampled audio
    """
    if use_cache:
        return _get_num_samples(name)
    return _get_num_samples.__wrapped__(name)


@lru_cache
def _get_num_samples(name: str | Path) -> int:
    """Get the number of samples resampled to the SonusAI sample rate in the given file

    :param name: File name
    :return: number of samples in resampled audio
    """
    import math

    import soundfile
    from pydub import AudioSegment

    from ..constants import SAMPLE_RATE
    from ..utils.tokenized_shell_vars import tokenized_expand

    expanded_name, _ = tokenized_expand(name)

    if expanded_name.endswith(".mp3"):
        sound = AudioSegment.from_mp3(expanded_name)
        samples = sound.frame_count()
        sample_rate = sound.frame_rate
    elif expanded_name.endswith(".m4a"):
        sound = AudioSegment.from_file(expanded_name)
        samples = sound.frame_count()
        sample_rate = sound.frame_rate
    else:
        info = soundfile.info(expanded_name)
        samples = info.frames
        sample_rate = info.samplerate

    return math.ceil(SAMPLE_RATE * samples / sample_rate)



from importlib.resources import as_file
from importlib.resources import files

SAMPLE_RATE = 16000
CHANNEL_COUNT = 1
BIT_DEPTH = 32
SAMPLE_BYTES = BIT_DEPTH // 8
FLOAT_BYTES = 4

with as_file(files("sonusai.data").joinpath("whitenoise.wav")) as path:
    DEFAULT_NOISE = str(path)



from typing import Any


def _get_hdf5_name(location: str, index: str) -> str:
    from os.path import join

    return join(location, index + ".h5")


def _get_pickle_name(location: str, index: str, item: str) -> str:
    from os.path import join

    return join(location, index, item + ".pkl")


def read_hdf5_data(location: str, index: str, items: list[str] | str) -> dict[str, Any]:
    """Read mixture, target, or noise data from an HDF5 file

    :param location: Location of the file
    :param index: Mixture, target, or noise index
    :param items: String(s) of data to retrieve
    :return: Dictionary of name: data
    """
    from os.path import exists
    from typing import Any

    import h5py
    import numpy as np

    def _get_dataset(file: h5py.File, d_name: str) -> Any:
        if d_name in file:
            data = np.array(file[d_name])
            if data.size == 1:
                item = data.item()
                if isinstance(item, bytes):
                    return item.decode("utf-8")
                return item
            return data
        return None

    if not isinstance(items, list):
        items = [items]

    h5_name = _get_hdf5_name(location, index)
    if exists(h5_name):
        try:
            with h5py.File(h5_name, "r") as f:
                result = {item: _get_dataset(f, item) for item in items}
        except Exception as e:
            raise OSError(f"Error reading {h5_name}: {e}") from e
    else:
        result = {item: None for item in items}

    return result


def write_hdf5_data(location: str, index: str, items: dict[str, Any]) -> None:
    """Write mixture, target, or noise data to an HDF5 file

    :param location: Location of the file
    :param index: Mixture, target, or noise index
    :param items: Dictionary of name: data
    """
    import h5py

    h5_name = _get_hdf5_name(location, index)
    with h5py.File(h5_name, "a") as f:
        for name, data in items.items():
            if name in f:
                del f[name]
            f.create_dataset(name=name, data=data)


def read_pickle_data(location: str, index: str, items: list[str] | str) -> dict[str, Any]:
    """Read mixture, target, or noise data from a pickle file

    :param location: Location of the file
    :param index: Mixture, target, or noise index
    :param items: String(s) of data to retrieve
    :return: Dictionary of name: data
    """
    import pickle
    from os.path import exists
    from typing import Any

    if not isinstance(items, list):
        items = [items]

    result: dict[str, Any] = {}
    for item in items:
        pkl_name = _get_pickle_name(location, index, item)
        if exists(pkl_name):
            with open(pkl_name, "rb") as f:
                result[item] = pickle.load(f)  # noqa: S301
        else:
            result[item] = None

    return result


def write_pickle_data(location: str, index: str, items: dict[str, Any]) -> None:
    """Write mixture, target, or noise data to a pickle file

    :param location: Location of the file
    :param index: Mixture, target, or noise index
    :param items: Dictionary of name: data
    """
    import pickle
    from os import makedirs
    from os.path import join

    directory = join(location, index)
    makedirs(directory, exist_ok=True)
    for name, data in items.items():
        pkl_name = _get_pickle_name(location, index, name)
        with open(pkl_name, "wb") as f:
            f.write(pickle.dumps(data))


def clear_pickle_data(location: str, index: str, items: list[str] | str) -> None:
    """Clear mixture, target, or noise data pickle file

    :param location: Location of the file
    :param index: Mixture, target, or noise index
    :param items: String(s) of data to retrieve
    """
    from pathlib import Path

    if not isinstance(items, list):
        items = [items]

    for item in items:
        Path(_get_pickle_name(location, index, item)).unlink(missing_ok=True)


def read_cached_data(location: str, name: str, index: str, items: list[str] | str) -> dict[str, Any]:
    """Read cached data from a file

    :param location: Location of the mixture database
    :param name: Data name ('mixture', 'target', or 'noise')
    :param index: Data index (mixture, target, or noise ID)
    :param items: String(s) of data to retrieve
    :return: Data (or tuple of data)
    """
    from os.path import join

    return read_pickle_data(join(location, name), index, items)


def write_cached_data(location: str, name: str, index: str, items: dict[str, Any]) -> None:
    """Write data to a file

    :param location: Location of the mixture database
    :param name: Data name ('mixture', 'target', or 'noise')
    :param index: Data index (mixture, target, or noise ID)
    :param items: Dictionary of name: data
    """
    from os.path import join

    write_pickle_data(join(location, name), index, items)


def clear_cached_data(location: str, name: str, index: str, items: list[str] | str) -> None:
    """Remove cached data file(s)

    :param location: Location of the mixture database
    :param name: Data name ('mixture', 'target', or 'noise')
    :param index: Data index (mixture, target, or noise ID)
    :param items: String(s) of data to clear
    """
    from os.path import join

    clear_pickle_data(join(location, name), index, items)



def _load_yaml(name: str) -> dict:
    """Load YAML file

    :param name: File name
    :return: Dictionary of config data
    """
    import yaml

    with open(file=name) as f:
        config = yaml.safe_load(f)

    return config


def _default_config() -> dict:
    """Load default SonusAI config

    :return: Dictionary of default config data
    """
    from .constants import DEFAULT_CONFIG

    try:
        return _load_yaml(DEFAULT_CONFIG)
    except Exception as e:
        raise OSError(f"Error loading default config: {e}") from e


def _update_config_from_file(filename: str, given_config: dict) -> dict:
    """Update the given config with the config in the specified YAML file

    :param filename: File name
    :param given_config: Config dictionary to update
    :return: Updated config dictionary
    """
    from copy import deepcopy

    updated_config = deepcopy(given_config)

    try:
        file_config = _load_yaml(filename)
    except Exception as e:
        raise OSError(f"Error loading config from {filename}: {e}") from e

    # Use default config as base and overwrite with given config keys as found
    if file_config:
        for key in updated_config:
            if key in file_config:
                updated_config[key] = file_config[key]

    return updated_config


def load_config(name: str) -> dict:
    """Load SonusAI default config and update with given location (performing SonusAI variable substitution)

    :param name: Directory containing mixture database
    :return: Dictionary of config data
    """
    from os.path import join

    return _update_config_from_file(filename=join(name, "config.yml"), given_config=_default_config())



from functools import lru_cache
from typing import Any


def load_object(name: str, use_cache: bool = True) -> Any:
    """Load an object from a pickle file"""
    if use_cache:
        return _load_object(name)
    return _load_object.__wrapped__(name)


@lru_cache
def _load_object(name: str) -> Any:
    import pickle
    from os.path import exists

    if exists(name):
        with open(name, "rb") as f:
            return pickle.load(f)  # noqa: S301

    raise FileNotFoundError(name)



import numpy as np

from ..constants import SAMPLE_RATE


def calc_pesq(
    hypothesis: np.ndarray,
    reference: np.ndarray,
    error_value: float = 0.0,
    sample_rate: int = SAMPLE_RATE,
) -> float:
    """Computes the PESQ score of hypothesis vs. reference

    Upon error, assigns a value of 0, or user specified value in error_value

    :param hypothesis: estimated audio
    :param reference: reference audio
    :param error_value: value to use if error occurs
    :param sample_rate: sample rate of audio
    :return: value between -0.5 to 4.5
    """
    import warnings

    from pesq import pesq

    from .. import logger

    try:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            score = pesq(fs=sample_rate, ref=reference, deg=hypothesis, mode="wb")
    except Exception as e:
        logger.debug(f"PESQ error {e}")
        score = error_value

    return score



def get_label_names(num_labels: int, file: str | None = None) -> list:
    """Return label names in a list. Read from CSV file, if provided."""
    import csv

    if file is None:
        return [f"Class {val + 1}" for val in range(num_labels)]

    label_names = [""] * num_labels
    with open(file) as f:
        reader = csv.DictReader(f)
        if reader.fieldnames is None or "index" not in reader.fieldnames or "display_name" not in reader.fieldnames:
            raise ValueError("Missing required fields in labels CSV.")

        for row in reader:
            index = int(row["index"]) - 1
            if index >= num_labels:
                raise ValueError("The number of given label names does not match the number of labels.")
            label_names[index] = row["display_name"]

    return label_names



from collections.abc import Sequence
from typing import Any


def dataclass_from_dict(klass, dikt: dict) -> Any:
    """Convert dictionary to dataclass."""
    try:
        field_types = klass.__annotations__
        return klass(**{f: dataclass_from_dict(field_types[f], dikt[f]) for f in dikt})
    except AttributeError:
        return dikt


def list_dataclass_from_dict(klass, dikt: Sequence[dict]) -> list[Any]:
    """Convert list of dictionary to list of dataclass."""
    return [dataclass_from_dict(klass.__args__[0], f) for f in dikt]


def original_dataclass_from_dict(klass, dikt):
    """Convert dictionary to dataclass."""
    try:
        field_types = klass.__annotations__
        return klass(**{f: dataclass_from_dict(field_types[f], dikt[f]) for f in dikt})
    except AttributeError:
        if isinstance(dikt, tuple | list):
            return [dataclass_from_dict(klass.__args__[0], f) for f in dikt]
        return dikt



"""sonusai genft

usage: genft [-hvsn] [-i MIXID] LOC

options:
    -h, --help
    -v, --verbose               Be verbose.
    -i MIXID, --mixid MIXID     Mixture ID(s) to generate. [default: *].
    -s, --segsnr                Save segsnr. [default: False].
    -n, --nopar                 Do not run in parallel. [default: False].

Generate SonusAI feature/truth data from a SonusAI mixture database.

Inputs:
    LOC         A SonusAI mixture database directory.
    MIXID       A glob of mixture ID(s) to generate.

Outputs the following to the mixture database directory:
    <id>
        feature.pkl
        truth_f.pkl
        segsnr.pkl (optional)
        metadata.txt
    genft.log

"""

from sonusai.datatypes import GeneralizedIDs
from sonusai.datatypes import GenFTData


def genft(
    location: str,
    mixids: GeneralizedIDs = "*",
    compute_truth: bool = True,
    compute_segsnr: bool = False,
    cache: bool = False,
    show_progress: bool = False,
    force: bool = True,
    no_par: bool = False,
) -> list[GenFTData]:
    from functools import partial

    from sonusai.mixture import MixtureDatabase
    from sonusai.utils import par_track
    from sonusai.utils import track

    mixdb = MixtureDatabase(location)
    mixids = mixdb.mixids_to_list(mixids)

    progress = track(total=len(mixids), disable=not show_progress)
    results = par_track(
        partial(
            _genft_kernel,
            location=location,
            compute_truth=compute_truth,
            compute_segsnr=compute_segsnr,
            cache=cache,
            force=force,
        ),
        mixids,
        progress=progress,
        no_par=no_par,
    )
    progress.close()

    return results


def _genft_kernel(
    m_id: int,
    location: str,
    compute_truth: bool,
    compute_segsnr: bool,
    cache: bool,
    force: bool,
) -> GenFTData:
    from functools import partial
    from typing import Any

    from sonusai.mixture import MixtureDatabase
    from sonusai.mixture import write_cached_data
    from sonusai.mixture import write_mixture_metadata

    mixdb = MixtureDatabase(location)

    write_func = partial(write_cached_data, location=mixdb.location, name="mixture", index=mixdb.mixture(m_id).name)

    result = GenFTData()

    mixture = mixdb.mixture_mixture(m_id)

    feature, truth_f = mixdb.mixture_ft(m_id=m_id, mixture=mixture, force=force)
    result.feature = feature
    items: dict[str, Any] = {"feature": feature}

    if compute_truth:
        result.truth_f = truth_f
        items["truth_f"] = truth_f

    if compute_segsnr:
        segsnr_t = mixdb.mixture_segsnr_t(m_id)
        result.segsnr = mixdb.mixture_segsnr(m_id=m_id, segsnr_t=segsnr_t, force=force, cache=cache)

    if cache:
        write_func(items=items)
        write_mixture_metadata(mixdb, m_id=m_id)

    return result


def main() -> None:
    from docopt import docopt

    from sonusai import __version__ as sai_version
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sai_version, options_first=True)

    import time
    from os.path import join

    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import logger
    from sonusai import update_console_handler
    from sonusai.constants import SAMPLE_RATE
    from sonusai.mixture import MixtureDatabase
    from sonusai.mixture import check_audio_files_exist
    from sonusai.utils import human_readable_size
    from sonusai.utils import seconds_to_hms

    verbose = args["--verbose"]
    mixids = args["--mixid"]
    compute_segsnr = args["--segsnr"]
    no_par = args["--nopar"]
    location = args["LOC"]

    start_time = time.monotonic()

    create_file_handler(join(location, "genft.log"))
    update_console_handler(verbose)
    initial_log_messages("genft")

    logger.info(f"Load mixture database from {location}")
    mixdb = MixtureDatabase(location)
    mixids = mixdb.mixids_to_list(mixids)

    total_samples = mixdb.total_samples(mixids)
    duration = total_samples / SAMPLE_RATE
    total_transform_frames = total_samples // mixdb.ft_config.overlap
    total_feature_frames = total_samples // mixdb.feature_step_samples

    logger.info("")
    logger.info(f"Found {len(mixids):,} mixtures to process")
    logger.info(
        f"{total_samples:,} samples, "
        f"{total_transform_frames:,} transform frames, "
        f"{total_feature_frames:,} feature frames"
    )

    check_audio_files_exist(mixdb)

    genft(
        location=location,
        mixids=mixids,
        compute_segsnr=compute_segsnr,
        cache=True,
        show_progress=True,
        no_par=no_par,
    )

    logger.info(f"Wrote {len(mixids)} mixtures to {location}")
    logger.info("")
    logger.info(f"Duration: {seconds_to_hms(seconds=duration)}")
    logger.info(
        f"feature:  {human_readable_size(total_feature_frames * mixdb.fg_stride * mixdb.feature_parameters * 4, 1)}"
    )
    logger.info(f"truth_f:  {human_readable_size(total_feature_frames * mixdb.num_classes * 4, 1)}")
    if compute_segsnr:
        logger.info(f"segsnr:   {human_readable_size(total_transform_frames * 4, 1)}")

    end_time = time.monotonic()
    logger.info(f"Completed in {seconds_to_hms(seconds=end_time - start_time)}")
    logger.info("")


if __name__ == "__main__":
    from sonusai import exception_handler
    from sonusai.utils import register_keyboard_interrupt

    register_keyboard_interrupt()
    try:
        main()
    except Exception as e:
        exception_handler(e)



def calculate_input_shape(feature: str, flatten: bool = False, timesteps: int = 0, add1ch: bool = False) -> list[int]:
    """
    Calculate input shape given feature and user-specified reshape parameters.

    Inputs:
        feature:     String defining the Aaware feature used in SonusAI, typically  mixdb.feature.
        flatten:     If true, flatten the 2D spectrogram from SxB to S*B.
        timesteps:   Pre-pend timesteps dimension if non-zero, size = timesteps.
        add1ch:      Append channel dimension of size 1, (channel last).
    """
    from pyaaware import FeatureGenerator

    fg = FeatureGenerator(feature_mode=feature)

    if flatten:
        in_shape = [fg.stride * fg.feature_parameters]
    else:
        in_shape = [fg.stride, fg.feature_parameters]

    if timesteps > 0:
        in_shape.insert(0, timesteps)

    if add1ch:
        in_shape.append(1)

    return in_shape



from ..datatypes import AudioT


def pad_audio_to_frame(audio: AudioT, frame_length: int = 1) -> AudioT:
    """Pad audio to be a multiple of frame length

    :param audio: Audio
    :param frame_length: Pad resulting audio to be a multiple of this
    :return: Padded audio
    """
    return pad_audio_to_length(audio, get_padded_length(len(audio), frame_length))


def get_padded_length(length: int, frame_length: int) -> int:
    """Get the number of pad samples needed

    :param length: Length of audio
    :param frame_length: Desired length will be a multiple of this
    :return: Padded length
    """
    mod = int(length % frame_length)
    pad_length = frame_length - mod if mod else 0
    return length + pad_length


def pad_audio_to_length(audio: AudioT, length: int) -> AudioT:
    """Pad audio to given length

    :param audio: Audio
    :param length: Length of output
    :return: Padded audio
    """
    import numpy as np

    return np.pad(array=audio, pad_width=(0, length - len(audio)))



from pyaaware import ForwardTransform
from pyaaware import InverseTransform

from ..datatypes import AudioF
from ..datatypes import AudioT
from ..datatypes import EnergyT
from ..datatypes import FeatureGeneratorConfig
from ..datatypes import FeatureGeneratorInfo
from ..datatypes import GeneralizedIDs
from ..datatypes import Mixture
from ..datatypes import Source
from ..datatypes import Sources
from ..datatypes import SpeechMetadata
from ..datatypes import TransformConfig
from .db_datatypes import MixtureRecord
from .db_datatypes import SourceRecord
from .mixdb import MixtureDatabase


def generic_ids_to_list(num_ids: int, ids: GeneralizedIDs = "*") -> list[int]:
    """Resolve generalized IDs to a list of integers

    :param num_ids: Total number of indices
    :param ids: Generalized IDs
    :return: List of ID integers
    """
    all_ids = list(range(num_ids))

    if isinstance(ids, str):
        if ids == "*":
            return all_ids

        try:
            result = eval(f"{all_ids}[{ids}]")  # noqa: S307
            if isinstance(result, list):
                return result
            else:
                return [result]
        except NameError as e:
            raise ValueError(f"Empty ids {ids}: {e}") from e

    if isinstance(ids, range):
        result = list(ids)
    elif isinstance(ids, int):
        result = [ids]
    else:
        result = ids

    if not all(isinstance(x, int) and 0 <= x < num_ids for x in result):
        raise ValueError(f"Invalid entries in ids of {ids}")

    if not result:
        raise ValueError(f"Empty ids {ids}")

    return result


def get_feature_generator_info(fg_config: FeatureGeneratorConfig) -> FeatureGeneratorInfo:
    from pyaaware import FeatureGenerator

    from ..datatypes import TransformConfig

    fg = FeatureGenerator(feature_mode=fg_config.feature_mode)

    return FeatureGeneratorInfo(
        decimation=fg.decimation,
        stride=fg.stride,
        step=fg.step,
        feature_parameters=fg.feature_parameters,
        ft_config=TransformConfig(
            length=fg.ftransform_length,
            overlap=fg.ftransform_overlap,
            bin_start=fg.bin_start,
            bin_end=fg.bin_end,
            ttype=fg.ftransform_ttype,
        ),
        eft_config=TransformConfig(
            length=fg.eftransform_length,
            overlap=fg.eftransform_overlap,
            bin_start=fg.bin_start,
            bin_end=fg.bin_end,
            ttype=fg.eftransform_ttype,
        ),
        it_config=TransformConfig(
            length=fg.itransform_length,
            overlap=fg.itransform_overlap,
            bin_start=fg.bin_start,
            bin_end=fg.bin_end,
            ttype=fg.itransform_ttype,
        ),
    )


def mixture_all_speech_metadata(mixdb: MixtureDatabase, mixture: Mixture) -> dict[str, dict[str, SpeechMetadata]]:
    """Get a list of all speech metadata for the given mixture"""
    from praatio.utilities.constants import Interval

    from ..datatypes import SpeechMetadata

    results: dict[str, dict[str, SpeechMetadata]] = {}
    for category, source in mixture.all_sources.items():
        data: dict[str, SpeechMetadata] = {}
        for tier in mixdb.speaker_metadata_tiers:
            data[tier] = mixdb.speaker(mixdb.source_file(source.file_id).speaker_id, tier)

        for tier in mixdb.textgrid_metadata_tiers:
            item = get_textgrid_tier_from_source_file(mixdb.source_file(source.file_id).name, tier)
            if isinstance(item, list):
                # Check for tempo effect and adjust Interval start and end data as needed
                entries = []
                for entry in item:
                    entries.append(
                        Interval(
                            entry.start / source.pre_tempo,
                            entry.end / source.pre_tempo,
                            entry.label,
                        )
                    )
                data[tier] = entries
            else:
                data[tier] = item
        results[category] = data

    return results


def mixture_metadata(mixdb: MixtureDatabase, m_id: int | None = None, mixture: Mixture | None = None) -> str:
    """Create a string of metadata for a Mixture

    :param mixdb: Mixture database
    :param m_id: Mixture ID
    :param mixture: Mixture record
    :return: String of metadata
    """
    if m_id is not None:
        mixture = mixdb.mixture(m_id)

    if mixture is None:
        raise ValueError("No mixture specified.")

    metadata = ""
    speech_metadata = mixture_all_speech_metadata(mixdb, mixture)
    metadata += f"samples: {mixture.samples}\n"
    for category, source in mixture.all_sources.items():
        source_file = mixdb.source_file(source.file_id)
        metadata += f"{category} name: {source_file.name}\n"
        metadata += f"{category} effects: {source.effects.to_dict()}\n"
        metadata += f"{category} pre_tempo: {source.pre_tempo}\n"
        metadata += f"{category} class indices: {source_file.class_indices}\n"
        metadata += f"{category} start: {source.start}\n"
        metadata += f"{category} repeat: {source.repeat}\n"
        metadata += f"{category} snr: {source.snr}\n"
        metadata += f"{category} random_snr: {source.snr.is_random}\n"
        metadata += f"{category} snr_gain: {source.snr_gain}\n"
        for key in source_file.truth_configs:
            metadata += f"{category} truth '{key}' function: {source_file.truth_configs[key].function}\n"
            metadata += f"{category} truth '{key}' config:   {source_file.truth_configs[key].config}\n"
        for key in speech_metadata[category]:
            metadata += f"{category} speech {key}: {speech_metadata[category][key]}\n"

    return metadata


def write_mixture_metadata(mixdb: MixtureDatabase, m_id: int | None = None, mixture: Mixture | None = None) -> None:
    """Write mixture metadata to a text file

    :param mixdb: Mixture database
    :param m_id: Mixture ID
    :param mixture: Mixture record
    """
    from os.path import join

    if m_id is not None:
        name = mixdb.mixture(m_id).name
    elif mixture is not None:
        name = mixture.name
    else:
        raise ValueError("No mixture specified.")

    name = join(mixdb.location, "mixture", name, "metadata.txt")
    with open(file=name, mode="w") as f:
        f.write(mixture_metadata(mixdb, m_id, mixture))


def from_mixture(mixture: Mixture) -> tuple[str, int, int, int]:
    return mixture.name, mixture.samples, mixture.spectral_mask_id, mixture.spectral_mask_seed


def to_mixture(entry: MixtureRecord, sources: Sources) -> Mixture:
    return Mixture(
        name=entry.name,
        samples=entry.samples,
        all_sources=sources,
        spectral_mask_id=entry.spectral_mask_id,
        spectral_mask_seed=entry.spectral_mask_seed,
    )


def from_source(source: Source) -> tuple[str, int, float, bool, float, float, bool, int]:
    return (
        source.effects.to_json(),
        source.file_id,
        source.pre_tempo,
        source.repeat,
        source.snr,
        source.snr_gain,
        source.snr.is_random,
        source.start,
    )


def to_source(entry: SourceRecord) -> Source:
    import json

    from ..datatypes import Effects
    from ..datatypes import UniversalSNR
    from ..utils.dataclass_from_dict import dataclass_from_dict

    return Source(
        file_id=entry.file_id,
        effects=dataclass_from_dict(Effects, json.loads(entry.effects)),
        start=entry.start,
        repeat=entry.repeat,
        snr=UniversalSNR(entry.snr, entry.snr_random),
        snr_gain=entry.snr_gain,
        pre_tempo=entry.pre_tempo,
    )


def get_transform_from_audio(audio: AudioT, transform: ForwardTransform) -> tuple[AudioF, EnergyT]:
    """Apply forward transform to input audio data to generate transform data

    :param audio: Time domain data [samples]
    :param transform: ForwardTransform object
    :return: Frequency domain data [frames, bins], Energy [frames]
    """
    import torch

    f, e = transform.execute_all(torch.from_numpy(audio))

    return f.numpy(), e.numpy()


def forward_transform(audio: AudioT, config: TransformConfig) -> AudioF:
    """Transform time domain data into frequency domain using the forward transform config from the feature

    A new transform is used for each call; i.e., state is not maintained between calls to forward_transform().

    :param audio: Time domain data [samples]
    :param config: Transform configuration
    :return: Frequency domain data [frames, bins]
    """
    from pyaaware import ForwardTransform

    audio_f, _ = get_transform_from_audio(
        audio=audio,
        transform=ForwardTransform(
            length=config.length,
            overlap=config.overlap,
            bin_start=config.bin_start,
            bin_end=config.bin_end,
            ttype=config.ttype,
        ),
    )
    return audio_f


def get_audio_from_transform(data: AudioF, transform: InverseTransform) -> tuple[AudioT, EnergyT]:
    """Apply inverse transform to input transform data to generate audio data

    :param data: Frequency domain data [frames, bins]
    :param transform: InverseTransform object
    :return: Time domain data [samples], Energy [frames]
    """

    import torch

    t, e = transform.execute_all(torch.from_numpy(data))

    return t.numpy(), e.numpy()


def inverse_transform(transform: AudioF, config: TransformConfig) -> AudioT:
    """Transform frequency domain data into time domain using the inverse transform config from the feature

    A new transform is used for each call; i.e., state is not maintained between calls to inverse_transform().

    :param transform: Frequency domain data [frames, bins]
    :param config: Transform configuration
    :return: Time domain data [samples]
    """
    from pyaaware import InverseTransform

    audio, _ = get_audio_from_transform(
        data=transform,
        transform=InverseTransform(
            length=config.length,
            overlap=config.overlap,
            bin_start=config.bin_start,
            bin_end=config.bin_end,
            ttype=config.ttype,
            gain=1,
        ),
    )
    return audio


def check_audio_files_exist(mixdb: MixtureDatabase) -> None:
    """Walk through all the noise and target audio files in a mixture database ensuring that they exist"""
    from os.path import exists

    from ..utils.tokenized_shell_vars import tokenized_expand

    for source_files in mixdb.source_files.values():
        for source_file in source_files:
            file_name, _ = tokenized_expand(source_file.name)
            if not exists(file_name):
                raise OSError(f"Could not find {file_name}")


def get_textgrid_tier_from_source_file(source_file: str, tier: str) -> SpeechMetadata | None:
    from pathlib import Path

    from praatio import textgrid
    from praatio.utilities.constants import Interval

    from ..utils.tokenized_shell_vars import tokenized_expand

    textgrid_file = Path(tokenized_expand(source_file)[0]).with_suffix(".TextGrid")
    if not textgrid_file.exists():
        return None

    tg = textgrid.openTextgrid(str(textgrid_file), includeEmptyIntervals=False)

    if tier not in tg.tierNames:
        return None

    entries = tg.getTier(tier).entries
    if len(entries) > 1:
        return [entry for entry in entries if isinstance(entry, Interval)]

    if len(entries) == 1:
        return entries[0].label

    return None


def frames_from_samples(samples: int, step_samples: int) -> int:
    import numpy as np

    return int(np.ceil(samples / step_samples))



from pathlib import Path


def tokenized_expand(name: str | bytes | Path) -> tuple[str, dict[str, str]]:
    """Expand shell variables of the forms $var, ${var} and %var%.
    Unknown variables are left unchanged.

    Expand paths containing shell variable substitutions. The following rules apply:
        - no expansion within single quotes
        - '$$' is translated into '$'
        - '%%' is translated into '%' if '%%' are not seen in %var1%%var2%
        - ${var} is accepted.
        - $varname is accepted.
        - %var% is accepted.
        - vars can be made out of letters, digits and the characters '_-'
        (though is not verified in the ${var} and %var% cases)

    :param name: String to expand
    :return: Tuple of (expanded string, dictionary of tokens)
    """
    import os
    import string

    from ..constants import DEFAULT_NOISE

    os.environ["default_noise"] = str(DEFAULT_NOISE)  # noqa: SIM112

    if isinstance(name, bytes):
        name = name.decode("utf-8")

    if isinstance(name, Path):
        name = name.as_posix()

    name = os.fspath(name)
    token_map: dict = {}

    if "$" not in name and "%" not in name:
        return name, token_map

    var_chars = string.ascii_letters + string.digits + "_-"
    quote = "'"
    percent = "%"
    brace = "{"
    rbrace = "}"
    dollar = "$"
    environ = os.environ

    result = name[:0]
    index = 0
    path_len = len(name)
    while index < path_len:
        c = name[index : index + 1]
        if c == quote:  # no expansion within single quotes
            name = name[index + 1 :]
            path_len = len(name)
            try:
                index = name.index(c)
                result += c + name[: index + 1]
            except ValueError:
                result += c + name
                index = path_len - 1
        elif c == percent:  # variable or '%'
            if name[index + 1 : index + 2] == percent:
                result += c
                index += 1
            else:
                name = name[index + 1 :]
                path_len = len(name)
                try:
                    index = name.index(percent)
                except ValueError:
                    result += percent + name
                    index = path_len - 1
                else:
                    var = name[:index]
                    try:
                        if environ is None:
                            value = os.fsencode(os.environ[os.fsdecode(var)]).decode("utf-8")  # type: ignore[unreachable]
                        else:
                            value = environ[var]
                        token_map[var] = value
                    except KeyError:
                        value = percent + var + percent
                    result += value
        elif c == dollar:  # variable or '$$'
            if name[index + 1 : index + 2] == dollar:
                result += c
                index += 1
            elif name[index + 1 : index + 2] == brace:
                name = name[index + 2 :]
                path_len = len(name)
                try:
                    index = name.index(rbrace)
                except ValueError:
                    result += dollar + brace + name
                    index = path_len - 1
                else:
                    var = name[:index]
                    try:
                        if environ is None:
                            value = os.fsencode(os.environ[os.fsdecode(var)]).decode("utf-8")  # type: ignore[unreachable]
                        else:
                            value = environ[var]
                        token_map[var] = value
                    except KeyError:
                        value = dollar + brace + var + rbrace
                    result += value
            else:
                var = name[:0]
                index += 1
                c = name[index : index + 1]
                while c and c in var_chars:
                    var += c
                    index += 1
                    c = name[index : index + 1]
                try:
                    if environ is None:
                        value = os.fsencode(os.environ[os.fsdecode(var)]).decode("utf-8")  # type: ignore[unreachable]
                    else:
                        value = environ[var]
                    token_map[var] = value
                except KeyError:
                    value = dollar + var
                result += value
                if c:
                    index -= 1
        else:
            result += c
        index += 1

    return result, token_map


def tokenized_replace(name: str, tokens: dict[str, str]) -> str:
    """Replace text with shell variables.

    :param name: String to replace
    :param tokens: Dictionary of replacement tokens
    :return: replaced string
    """
    for key, value in tokens.items():
        name = name.replace(value, f"${key}")
    return name



import numpy as np


def get_ir_delay(file: str) -> int:
    from ..utils.rand import seed_context
    from .audio import raw_read_audio

    ir, sample_rate = raw_read_audio(file)

    with seed_context(42):
        wgn_ref = np.random.normal(loc=0, scale=0.2, size=int(np.ceil(0.05 * sample_rate))).astype(np.float32)

    wgn_conv = np.convolve(ir, wgn_ref)

    return int(np.round(tdoa(wgn_conv, wgn_ref, interp=16, phat=True)))


def tdoa(signal: np.ndarray, reference: np.ndarray, interp: int = 1, phat: bool = False, fs: int | float = 1) -> float:
    """Estimates the shift of array signal with respect to reference using generalized cross-correlation.

    :param signal: The array whose tdoa is measured
    :param reference: The reference array
    :param interp: Interpolation factor for the output array
    :param phat: Apply the PHAT weighting
    :param fs: The sampling frequency of the input arrays
    :return: The estimated delay between the two arrays
    """
    n_reference = reference.shape[0]

    r_12 = correlate(signal, reference, interp=interp, phat=phat)

    delay = (np.argmax(np.abs(r_12)) / interp - (n_reference - 1)) / fs

    return float(delay)


def correlate(x1: np.ndarray, x2: np.ndarray, interp: int = 1, phat: bool = False) -> np.ndarray:
    """Compute the cross-correlation between x1 and x2

    :param x1: Input array 1
    :param x2: Input array 2
    :param interp: Interpolation factor for the output array
    :param phat: Apply the PHAT weighting
    :return: The cross-correlation between the two arrays
    """
    n_x1 = x1.shape[0]
    n_x2 = x2.shape[0]

    n = n_x1 + n_x2 - 1

    fft1 = np.fft.rfft(x1, n=n)
    fft2 = np.fft.rfft(x2, n=n)

    if phat:
        eps1 = np.mean(np.abs(fft1)) * 1e-10
        fft1 /= np.abs(fft1) + eps1
        eps2 = np.mean(np.abs(fft2)) * 1e-10
        fft2 /= np.abs(fft2) + eps2

    out = np.fft.irfft(fft1 * np.conj(fft2), n=int(n * interp))

    return np.concatenate([out[-interp * (n_x2 - 1) :], out[: (interp * n_x1)]])



from ...datatypes import Truth
from ..mixdb import MixtureDatabase


def metadata_validate(config: dict) -> None:
    if len(config) == 0:
        raise AttributeError("metadata truth function is missing config")

    parameters = ["tier"]
    for parameter in parameters:
        if parameter not in config:
            raise AttributeError(f"metadata truth function is missing required '{parameter}'")


def metadata_parameters(_feature: str, _num_classes: int, _config: dict) -> int | None:
    return None


def metadata(mixdb: MixtureDatabase, m_id: int, category: str, config: dict) -> Truth:
    """Metadata truth generation function

    Retrieves metadata from target.
    """
    return mixdb.mixture_speech_metadata(m_id, config["tier"])[category]



from importlib.resources import as_file
from importlib.resources import files

REQUIRED_TRUTH_CONFIG_FIELDS = ["function", "stride_reduction"]
REQUIRED_ASR_CONFIG_FIELDS = ["engine"]

with as_file(files("sonusai.config").joinpath("config.yml")) as path:
    DEFAULT_CONFIG = str(path)



import warnings
from collections.abc import Callable
from collections.abc import Iterable
from multiprocessing import current_process
from multiprocessing import get_context
from typing import Any

from tqdm import TqdmExperimentalWarning
from tqdm.rich import tqdm

warnings.filterwarnings(action="ignore", category=TqdmExperimentalWarning)

track = tqdm

CONTEXT = "fork"


def par_track(
    func: Callable,
    *iterables: Iterable,
    initializer: Callable[..., None] | None = None,
    initargs: Iterable[Any] | None = None,
    progress: tqdm | None = None,
    num_cpus: int | float | None = None,
    total: int | None = None,
    no_par: bool = False,
) -> list[Any]:
    """Performs a parallel ordered imap with tqdm progress."""
    from collections.abc import Sized

    from psutil import cpu_count

    if total is None:
        _total = min(len(iterable) for iterable in iterables if isinstance(iterable, Sized))
    else:
        _total = int(total)

    results: list[Any] = [None] * _total
    if no_par or current_process().daemon:
        if initializer is not None:
            if initargs is not None:
                initializer(*initargs)
            else:
                initializer()

        for n, result in enumerate(map(func, *iterables)):
            results[n] = result
            if progress is not None:
                progress.update()
    else:
        if num_cpus is None:
            _num_cpus = max(cpu_count() - 2, 1)
        elif isinstance(num_cpus, float):
            _num_cpus = int(round(num_cpus * cpu_count()))
        else:
            _num_cpus = int(num_cpus)

        _num_cpus = min(_num_cpus, _total)

        if initargs is None:
            initargs = []

        with get_context(CONTEXT).Pool(processes=_num_cpus, initializer=initializer, initargs=initargs) as pool:
            n = 0
            for result in pool.imap(func, *iterables):  # type: ignore[arg-type]
                results[n] = result
                n += 1
                if progress is not None:
                    progress.update()
            pool.close()
            pool.join()

    if progress is not None:
        progress.close()
    return results



from ..datatypes import AudioF


def power_compress(feature: AudioF) -> AudioF:
    import numpy as np

    mag = np.abs(feature)
    phase = np.angle(feature)
    mag = mag**0.3
    real_compress = mag * np.cos(phase)
    imag_compress = mag * np.sin(phase)

    return real_compress + 1j * imag_compress


def power_uncompress(feature: AudioF) -> AudioF:
    import numpy as np

    mag = np.abs(feature)
    phase = np.angle(feature)
    mag = mag ** (1.0 / 0.3)
    real_uncompress = mag * np.cos(phase)
    imag_uncompress = mag * np.sin(phase)

    return real_uncompress + 1j * imag_uncompress



from numpy.lib.utils import source

from ...datatypes import Truth
from ..mixdb import MixtureDatabase


def sed_validate(config: dict) -> None:
    if len(config) == 0:
        raise AttributeError("sed truth function is missing config")

    parameters = ["thresholds"]
    for parameter in parameters:
        if parameter not in config:
            raise AttributeError(f"sed truth function is missing required '{parameter}'")

    thresholds = config["thresholds"]
    if not _strictly_decreasing(thresholds):
        raise ValueError(f"sed truth function 'thresholds' are not strictly decreasing: {thresholds}")


def sed_parameters(_feature: str, num_classes: int, _config: dict) -> int:
    return num_classes


def sed(mixdb: MixtureDatabase, m_id: int, category: str, config: dict) -> Truth:
    """Sound energy detection truth generation function

    Calculates sound energy detection truth using simple 3 threshold
    hysteresis algorithm. SED outputs 3 possible probabilities of
    sound presence: 1.0 present, 0.5 (transition/uncertain), 0 not
    present. The output values will be assigned to the truth output
    at the index specified in the config.

    Output shape: [:, num_classes]

    index       Truth index <int> or list(<int>)

    index indicates which truth fields should be set.
    0 indicates none, 1 is first element in truth output vector, 2 2nd element, etc.

                Examples:
                  index = 5       truth in class 5, truth(4, 1)
                  index = [1, 5]  truth in classes 1 and 5, truth([0, 4], 1)

                In mutually-exclusive mode, a frame is expected to only
                belong to one class and thus all probabilities must sum to
                1. This is effectively truth for a classifier with multichannel
                softmax output.

                For multi-label classification each class is an individual
                probability for that class and any given frame can be
                assigned to multiple classes/labels, i.e., the classes are
                not mutually-exclusive. For example, a NN classifier with
                multichannel sigmoid output. In this case, index could
                also be a vector with multiple class indices.
    """
    import numpy as np
    import torch
    from pyaaware import SED
    from pyaaware import ForwardTransform
    from pyaaware import feature_forward_transform_config
    from pyaaware import feature_inverse_transform_config

    source_audio = torch.from_numpy(mixdb.mixture_sources(m_id)[category])

    frame_size = feature_inverse_transform_config(mixdb.feature)["overlap"]

    ft = ForwardTransform(**feature_forward_transform_config(mixdb.feature))

    if len(source_audio) % frame_size != 0:
        raise ValueError(f"Number of samples in audio is not a multiple of {frame_size}")

    frames = ft.frames(source_audio)
    parameters = sed_parameters(mixdb.feature, mixdb.num_classes, config)
    if mixdb.mixture(m_id).all_sources[category].snr_gain == 0:
        return np.zeros((frames, parameters), dtype=np.float32)

    # SED wants 1-based indices
    s = SED(
        thresholds=config["thresholds"],
        index=mixdb.source_file(mixdb.mixture(m_id).all_sources[category].file_id).class_indices,
        frame_size=frame_size,
        num_classes=mixdb.num_classes,
    )

    # Compute energy
    target_energy = ft.execute_all(source_audio)[1].numpy()

    if frames != target_energy.shape[0]:
        raise ValueError("Incorrect frames calculation in sed truth function")

    return s.execute_all(target_energy)


def _strictly_decreasing(list_to_check: list) -> bool:
    from itertools import pairwise

    return all(x > y for x, y in pairwise(list_to_check))



# ruff: noqa: S608
from ..datatypes import AudioT
from ..datatypes import Effects
from ..datatypes import GenMixData
from ..datatypes import ImpulseResponseFile
from ..datatypes import Mixture
from ..datatypes import Source
from ..datatypes import SourceFile
from ..datatypes import SourcesAudioT
from ..datatypes import UniversalSNRGenerator
from .mixdb import MixtureDatabase


def config_file(location: str) -> str:
    from os.path import join

    return join(location, "config.yml")


def initialize_db(location: str, test: bool = False, verbose: bool = False) -> None:
    from .mixdb import db_connection

    con = db_connection(location=location, create=True, test=test, verbose=verbose)

    con.execute("""
    CREATE TABLE truth_config(
    id INTEGER PRIMARY KEY NOT NULL,
    config TEXT NOT NULL)
    """)

    con.execute("""
    CREATE TABLE truth_parameters(
    id INTEGER PRIMARY KEY NOT NULL,
    category TEXT NOT NULL,
    name TEXT NOT NULL,
    parameters INTEGER)
    """)

    con.execute("""
    CREATE TABLE source_file (
    id INTEGER PRIMARY KEY NOT NULL,
    category TEXT NOT NULL,
    class_indices TEXT,
    level_type TEXT NOT NULL,
    name TEXT NOT NULL,
    samples INTEGER NOT NULL,
    speaker_id INTEGER,
    FOREIGN KEY(speaker_id) REFERENCES speaker (id))
    """)

    con.execute("""
    CREATE TABLE ir_file (
    id INTEGER PRIMARY KEY NOT NULL,
    delay INTEGER NOT NULL,
    name TEXT NOT NULL)
    """)

    con.execute("""
    CREATE TABLE ir_tag (
    id INTEGER PRIMARY KEY NOT NULL,
    tag TEXT NOT NULL UNIQUE)
    """)

    con.execute("""
    CREATE TABLE ir_file_ir_tag (
    file_id INTEGER NOT NULL,
    tag_id INTEGER NOT NULL,
    FOREIGN KEY(file_id) REFERENCES ir_file (id),
    FOREIGN KEY(tag_id) REFERENCES ir_tag (id))
    """)

    con.execute("""
    CREATE TABLE speaker (
    id INTEGER PRIMARY KEY NOT NULL,
    parent TEXT NOT NULL)
    """)

    con.execute("""
    CREATE TABLE top (
    id INTEGER PRIMARY KEY NOT NULL,
    asr_configs TEXT NOT NULL,
    class_balancing BOOLEAN NOT NULL,
    feature TEXT NOT NULL,
    mixid_width INTEGER NOT NULL,
    num_classes INTEGER NOT NULL,
    seed INTEGER NOT NULL,
    speaker_metadata_tiers TEXT NOT NULL,
    textgrid_metadata_tiers TEXT NOT NULL,
    version INTEGER NOT NULL)
    """)

    con.execute("""
    CREATE TABLE class_label (
    id INTEGER PRIMARY KEY NOT NULL,
    label TEXT NOT NULL)
    """)

    con.execute("""
    CREATE TABLE class_weights_threshold (
    id INTEGER PRIMARY KEY NOT NULL,
    threshold FLOAT NOT NULL)
    """)

    con.execute("""
    CREATE TABLE spectral_mask (
    id INTEGER PRIMARY KEY NOT NULL,
    f_max_width INTEGER NOT NULL,
    f_num INTEGER NOT NULL,
    t_max_percent INTEGER NOT NULL,
    t_max_width INTEGER NOT NULL,
    t_num INTEGER NOT NULL)
    """)

    con.execute("""
    CREATE TABLE source_file_truth_config (
    source_file_id INTEGER NOT NULL,
    truth_config_id INTEGER NOT NULL,
    FOREIGN KEY(source_file_id) REFERENCES source_file (id),
    FOREIGN KEY(truth_config_id) REFERENCES truth_config (id))
    """)

    con.execute("""
    CREATE TABLE source (
    id INTEGER PRIMARY KEY NOT NULL,
    effects TEXT NOT NULL,
    file_id INTEGER NOT NULL,
    pre_tempo FLOAT NOT NULL,
    repeat BOOLEAN NOT NULL,
    snr FLOAT NOT NULL,
    snr_gain FLOAT NOT NULL,
    snr_random BOOLEAN NOT NULL,
    start INTEGER NOT NULL,
    UNIQUE(effects, file_id, pre_tempo, repeat, snr, snr_gain, snr_random, start),
    FOREIGN KEY(file_id) REFERENCES source_file (id))
    """)

    con.execute("""
    CREATE TABLE mixture (
    id INTEGER PRIMARY KEY NOT NULL,
    name TEXT NOT NULL,
    samples INTEGER NOT NULL,
    spectral_mask_id INTEGER NOT NULL,
    spectral_mask_seed INTEGER NOT NULL,
    FOREIGN KEY(spectral_mask_id) REFERENCES spectral_mask (id))
    """)

    con.execute("""
    CREATE TABLE mixture_source (
    mixture_id INTEGER NOT NULL,
    source_id INTEGER NOT NULL,
    FOREIGN KEY(mixture_id) REFERENCES mixture (id),
    FOREIGN KEY(source_id) REFERENCES source (id))
    """)

    con.commit()
    con.close()


def populate_top_table(location: str, config: dict, test: bool = False, verbose: bool = False) -> None:
    """Populate top table"""
    import json

    from .constants import MIXDB_VERSION
    from .mixdb import db_connection

    con = db_connection(location=location, readonly=False, test=test, verbose=verbose)
    con.execute(
        """
    INSERT INTO top (id, asr_configs, class_balancing, feature, mixid_width, num_classes,
    seed, speaker_metadata_tiers, textgrid_metadata_tiers, version)
    VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?)
    """,
        (
            1,
            json.dumps(config["asr_configs"]),
            config["class_balancing"],
            config["feature"],
            0,
            config["num_classes"],
            config["seed"],
            "",
            "",
            MIXDB_VERSION,
        ),
    )
    con.commit()
    con.close()


def populate_class_label_table(location: str, config: dict, test: bool = False, verbose: bool = False) -> None:
    """Populate class_label table"""
    from .mixdb import db_connection

    con = db_connection(location=location, readonly=False, test=test, verbose=verbose)
    con.executemany(
        "INSERT INTO class_label (label) VALUES (?)",
        [(item,) for item in config["class_labels"]],
    )
    con.commit()
    con.close()


def populate_class_weights_threshold_table(
    location: str,
    config: dict,
    test: bool = False,
    verbose: bool = False,
) -> None:
    """Populate class_weights_threshold table"""
    from .mixdb import db_connection

    class_weights_threshold = config["class_weights_threshold"]
    num_classes = config["num_classes"]

    if not isinstance(class_weights_threshold, list):
        class_weights_threshold = [class_weights_threshold]

    if len(class_weights_threshold) == 1:
        class_weights_threshold = [class_weights_threshold[0]] * num_classes

    if len(class_weights_threshold) != num_classes:
        raise ValueError(f"invalid class_weights_threshold length: {len(class_weights_threshold)}")

    con = db_connection(location=location, readonly=False, test=test, verbose=verbose)
    con.executemany(
        "INSERT INTO class_weights_threshold (threshold) VALUES (?)",
        [(item,) for item in class_weights_threshold],
    )
    con.commit()
    con.close()


def populate_spectral_mask_table(location: str, config: dict, test: bool = False, verbose: bool = False) -> None:
    """Populate spectral_mask table"""
    from .config import get_spectral_masks
    from .mixdb import db_connection

    con = db_connection(location=location, readonly=False, test=test, verbose=verbose)
    con.executemany(
        """
    INSERT INTO spectral_mask (f_max_width, f_num, t_max_percent, t_max_width, t_num) VALUES (?, ?, ?, ?, ?)
    """,
        [
            (
                item.f_max_width,
                item.f_num,
                item.t_max_percent,
                item.t_max_width,
                item.t_num,
            )
            for item in get_spectral_masks(config)
        ],
    )
    con.commit()
    con.close()


def populate_truth_parameters_table(location: str, config: dict, test: bool = False, verbose: bool = False) -> None:
    """Populate truth_parameters table"""
    from .config import get_truth_parameters
    from .mixdb import db_connection

    con = db_connection(location=location, readonly=False, test=test, verbose=verbose)
    con.executemany(
        """
    INSERT INTO truth_parameters (category, name, parameters) VALUES (?, ?, ?)
    """,
        [
            (
                item.category,
                item.name,
                item.parameters,
            )
            for item in get_truth_parameters(config)
        ],
    )
    con.commit()
    con.close()


def populate_source_file_table(
    location: str,
    files: list[SourceFile],
    test: bool = False,
    verbose: bool = False,
) -> None:
    """Populate source file table"""
    import json
    from pathlib import Path

    from .mixdb import db_connection

    _populate_truth_config_table(location, files, test, verbose)
    _populate_speaker_table(location, files, test, verbose)

    con = db_connection(location=location, readonly=False, test=test, verbose=verbose)

    cur = con.cursor()
    textgrid_metadata_tiers: set[str] = set()
    for file in files:
        # Get TextGrid tiers for source file and add to collection
        tiers = _get_textgrid_tiers_from_source_file(file.name)
        for tier in tiers:
            textgrid_metadata_tiers.add(tier)

        # Get truth settings for file
        truth_config_ids: list[int] = []
        if file.truth_configs:
            for name, config in file.truth_configs.items():
                ts = json.dumps({"name": name} | config.to_dict())
                cur.execute(
                    "SELECT truth_config.id FROM truth_config WHERE ? = truth_config.config",
                    (ts,),
                )
                truth_config_ids.append(cur.fetchone()[0])

        # Get speaker_id for source file
        cur.execute("SELECT speaker.id FROM speaker WHERE ? = speaker.parent", (Path(file.name).parent.as_posix(),))
        result = cur.fetchone()
        speaker_id = None
        if result is not None:
            speaker_id = result[0]

        # Add entry
        cur.execute(
            """
            INSERT INTO source_file (category, class_indices, level_type, name, samples, speaker_id)
            VALUES (?, ?, ?, ?, ?, ?)
            """,
            (
                file.category,
                json.dumps(file.class_indices),
                file.level_type,
                file.name,
                file.samples,
                speaker_id,
            ),
        )
        source_file_id = cur.lastrowid
        for truth_config_id in truth_config_ids:
            cur.execute(
                "INSERT INTO source_file_truth_config (source_file_id, truth_config_id) VALUES (?, ?)",
                (source_file_id, truth_config_id),
            )

    # Update textgrid_metadata_tiers in the top table
    con.execute(
        "UPDATE top SET textgrid_metadata_tiers=? WHERE ? = id", (json.dumps(sorted(textgrid_metadata_tiers)), 1)
    )

    con.commit()
    con.close()


def populate_impulse_response_file_table(
    location: str,
    files: list[ImpulseResponseFile],
    test: bool = False,
    verbose: bool = False,
) -> None:
    """Populate impulse response file table"""
    from .mixdb import db_connection

    _populate_impulse_response_tag_table(location, files, test, verbose)

    con = db_connection(location=location, readonly=False, test=test, verbose=verbose)

    cur = con.cursor()
    for file in files:
        # Get tags for file
        tag_ids: list[int] = []
        for tag in file.tags:
            cur.execute("SELECT id FROM ir_tag WHERE ? = tag", (tag,))
            tag_ids.append(cur.fetchone()[0])

        cur.execute("INSERT INTO ir_file (delay, name) VALUES (?, ?)", (file.delay, file.name))

        file_id = cur.lastrowid
        for tag_id in tag_ids:
            cur.execute("INSERT INTO ir_file_ir_tag (file_id, tag_id) VALUES (?, ?)", (file_id, tag_id))

    con.commit()
    con.close()


def update_mixid_width(location: str, num_mixtures: int, test: bool = False, verbose: bool = False) -> None:
    """Update the mixid width"""
    from ..utils.max_text_width import max_text_width
    from .mixdb import db_connection

    con = db_connection(location=location, readonly=False, test=test, verbose=verbose)
    con.execute("UPDATE top SET mixid_width=? WHERE ? = id", (max_text_width(num_mixtures), 1))
    con.commit()
    con.close()


def generate_mixtures(
    location: str,
    config: dict,
    effects: dict[str, list[Effects]],
    test: bool = False,
) -> list[Mixture]:
    """Generate mixtures"""
    mixdb = MixtureDatabase(location, test)

    effected_sources: dict[str, list[tuple[SourceFile, Effects]]] = {}
    for category in mixdb.source_files:
        effected_sources[category] = []
        for file in mixdb.source_files[category]:
            for effect in effects[category]:
                effected_sources[category].append((file, effect))

    mixtures: list[Mixture] = []
    for noise_mix_rule in config["sources"]["noise"]["mix_rules"]:
        match noise_mix_rule["mode"]:
            case "exhaustive":
                func = _exhaustive_noise_mix
            case "non-exhaustive":
                func = _non_exhaustive_noise_mix
            case "non-combinatorial":
                func = _non_combinatorial_noise_mix
            case _:
                raise ValueError(f"invalid noise mix_rule mode: {noise_mix_rule['mode']}")

        mixtures.extend(
            func(
                location=location,
                config=config,
                effected_sources=effected_sources,
                test=test,
            )
        )

    return mixtures


def populate_mixture_table(
    location: str,
    mixtures: list[Mixture],
    test: bool = False,
    verbose: bool = False,
    logging: bool = False,
    show_progress: bool = False,
) -> None:
    """Populate mixture table"""
    from .. import logger
    from ..utils.parallel import track
    from .helpers import from_mixture
    from .helpers import from_source
    from .mixdb import db_connection

    con = db_connection(location=location, readonly=False, test=test, verbose=verbose)

    # Populate source table
    if logging:
        logger.info("Populating mixture and source tables")
    for mixture in track(mixtures, disable=not show_progress):
        m_id = int(mixture.name) + 1
        con.execute(
            """
            INSERT INTO mixture (id, name, samples, spectral_mask_id, spectral_mask_seed)
            VALUES (?, ?, ?, ?, ?)
            """,
            (m_id, *from_mixture(mixture)),
        )

        for source in mixture.all_sources.values():
            con.execute(
                """
            INSERT OR IGNORE INTO source (effects, file_id, pre_tempo, repeat, snr, snr_gain, snr_random, start)
            VALUES (?, ?, ?, ?, ?, ?, ?, ?)
            """,
                from_source(source),
            )

            source_id = con.execute(
                """
                SELECT id
                FROM source
                WHERE ? = effects
                AND ? = file_id
                AND ? = pre_tempo
                AND ? = repeat
                AND ? = snr
                AND ? = snr_gain
                AND ? = snr_random
                AND ? = start
            """,
                from_source(source),
            ).fetchone()[0]
            con.execute("INSERT INTO mixture_source (mixture_id, source_id) VALUES (?, ?)", (m_id, source_id))

    if logging:
        logger.info("Closing mixture and source tables")
    con.commit()
    con.close()


def update_mixture(mixdb: MixtureDatabase, mixture: Mixture, with_data: bool = False) -> tuple[Mixture, GenMixData]:
    """Update mixture record with name, samples, and gains"""
    import numpy as np

    sources_audio: SourcesAudioT = {}
    post_audio: SourcesAudioT = {}
    for category in mixture.all_sources:
        mixture, sources_audio[category], post_audio[category] = _update_source(mixdb, mixture, category)

    mixture = _initialize_mixture_gains(mixdb, mixture, post_audio)

    mixture.name = f"{int(mixture.name):0{mixdb.mixid_width}}"

    if not with_data:
        return mixture, GenMixData()

    # Apply gains
    post_audio = {
        category: post_audio[category] * mixture.all_sources[category].snr_gain for category in mixture.all_sources
    }

    # Sum sources, noise, and mixture
    source_audio = np.sum([post_audio[category] for category in mixture.sources], axis=0)
    noise_audio = post_audio["noise"]
    mixture_audio = source_audio + noise_audio

    return mixture, GenMixData(
        sources=sources_audio,
        source=source_audio,
        noise=noise_audio,
        mixture=mixture_audio,
    )


def _update_source(mixdb: MixtureDatabase, mixture: Mixture, category: str) -> tuple[Mixture, AudioT, AudioT]:
    from .effects import apply_effects
    from .effects import conform_audio_to_length

    source = mixture.all_sources[category]
    org_audio = mixdb.read_source_audio(source.file_id)

    org_samples = len(org_audio)
    pre_audio = apply_effects(mixdb, org_audio, source.effects, pre=True, post=False)

    pre_samples = len(pre_audio)
    mixture.all_sources[category].pre_tempo = org_samples / pre_samples

    pre_audio = conform_audio_to_length(pre_audio, mixture.samples, source.repeat, source.start)

    post_audio = apply_effects(mixdb, pre_audio, source.effects, pre=False, post=True)
    if len(pre_audio) != len(post_audio):
        raise RuntimeError(f"post-truth effects changed length: {source.effects.post}")

    return mixture, pre_audio, post_audio


def _initialize_mixture_gains(mixdb: MixtureDatabase, mixture: Mixture, sources_audio: SourcesAudioT) -> Mixture:
    import numpy as np

    from ..utils.asl_p56 import asl_p56
    from ..utils.db import db_to_linear

    sources_energy: dict[str, float] = {}
    for category in mixture.all_sources:
        level_type = mixdb.source_file(mixture.all_sources[category].file_id).level_type
        match level_type:
            case "default":
                sources_energy[category] = float(np.mean(np.square(sources_audio[category])))
            case "speech":
                sources_energy[category] = asl_p56(sources_audio[category])
            case _:
                raise ValueError(f"Unknown level_type: {level_type}")

    # Initialize all gains to 1
    for category in mixture.all_sources:
        mixture.all_sources[category].snr_gain = 1

    # Resolve gains
    for category in mixture.all_sources:
        if mixture.is_noise_only and category != "noise":
            # Special case for zeroing out source data
            mixture.all_sources[category].snr_gain = 0
        elif mixture.is_source_only and category == "noise":
            # Special case for zeroing out noise data
            mixture.all_sources[category].snr_gain = 0
        elif category != "primary":
            if sources_energy["primary"] == 0 or sources_energy[category] == 0:
                # Avoid divide-by-zero
                mixture.all_sources[category].snr_gain = 1
            else:
                mixture.all_sources[category].snr_gain = float(
                    np.sqrt(sources_energy["primary"] / sources_energy[category])
                ) / db_to_linear(mixture.all_sources[category].snr)

    # Normalize gains
    max_snr_gain = max([source.snr_gain for source in mixture.all_sources.values()])
    for category in mixture.all_sources:
        mixture.all_sources[category].snr_gain = mixture.all_sources[category].snr_gain / max_snr_gain

    # Check for clipping in mixture
    mixture_audio = np.sum(
        [sources_audio[category] * mixture.all_sources[category].snr_gain for category in mixture.all_sources], axis=0
    )
    max_abs_audio = float(np.max(np.abs(mixture_audio)))
    clip_level = db_to_linear(-0.25)
    if max_abs_audio > clip_level:
        gain_adjustment = clip_level / max_abs_audio
        for category in mixture.all_sources:
            mixture.all_sources[category].snr_gain *= gain_adjustment

    # To improve repeatability, round results
    for category in mixture.all_sources:
        mixture.all_sources[category].snr_gain = round(mixture.all_sources[category].snr_gain, ndigits=5)

    return mixture


def _exhaustive_noise_mix(
    location: str,
    config: dict,
    effected_sources: dict[str, list[tuple[SourceFile, Effects]]],
    test: bool = False,
) -> list[Mixture]:
    """Use every noise/effect with every source/effect+interferences/effect"""
    from random import randint

    import numpy as np

    from ..datatypes import Mixture
    from ..datatypes import UniversalSNR
    from .effects import effects_from_rules
    from .effects import estimate_effected_length

    mixdb = MixtureDatabase(location, test)
    snrs = get_all_snrs_from_config(config)

    m_id = 0
    mixtures: list[Mixture] = []
    for noise_file, noise_rule in effected_sources["noise"]:
        noise_start = 0
        noise_effect = effects_from_rules(mixdb, noise_rule)
        noise_length = estimate_effected_length(noise_file.samples, noise_effect)

        for primary_file, primary_rule in effected_sources["primary"]:
            primary_effect = effects_from_rules(mixdb, primary_rule)
            primary_length = estimate_effected_length(primary_file.samples, primary_effect, mixdb.feature_step_samples)

            for spectral_mask_id in range(len(config["spectral_masks"])):
                for snr in snrs["noise"]:
                    mixtures.append(
                        Mixture(
                            name=str(m_id),
                            all_sources={
                                "primary": Source(
                                    file_id=primary_file.id,
                                    effects=primary_effect,
                                ),
                                "noise": Source(
                                    file_id=noise_file.id,
                                    effects=noise_effect,
                                    start=noise_start,
                                    repeat=True,
                                    snr=UniversalSNR(value=snr.value, is_random=snr.is_random),
                                ),
                            },
                            samples=primary_length,
                            spectral_mask_id=spectral_mask_id + 1,
                            spectral_mask_seed=randint(0, np.iinfo("i").max),  # noqa: S311
                        )
                    )
                    noise_start = int((noise_start + primary_length) % noise_length)
                    m_id += 1

    return mixtures


def _non_exhaustive_noise_mix(
    location: str,
    config: dict,
    effected_sources: dict[str, list[tuple[SourceFile, Effects]]],
    test: bool = False,
) -> list[Mixture]:
    """Cycle through every source/effect+interferences/effect without necessarily using all
    noise/effect combinations (reduced data set).
    """
    from random import randint

    import numpy as np

    from ..datatypes import Mixture
    from ..datatypes import UniversalSNR
    from .effects import effects_from_rules
    from .effects import estimate_effected_length

    mixdb = MixtureDatabase(location, test)
    snrs = get_all_snrs_from_config(config)

    next_noise = NextNoise(mixdb, effected_sources["noise"])

    m_id = 0
    mixtures: list[Mixture] = []
    for primary_file, primary_rule in effected_sources["primary"]:
        primary_effect = effects_from_rules(mixdb, primary_rule)
        primary_length = estimate_effected_length(primary_file.samples, primary_effect, mixdb.feature_step_samples)

        for spectral_mask_id in range(len(config["spectral_masks"])):
            for snr in snrs["noise"]:
                noise_file_id, noise_effect, noise_start = next_noise.generate(primary_file.samples)

                mixtures.append(
                    Mixture(
                        name=str(m_id),
                        all_sources={
                            "primary": Source(
                                file_id=primary_file.id,
                                effects=primary_effect,
                            ),
                            "noise": Source(
                                file_id=noise_file_id,
                                effects=noise_effect,
                                start=noise_start,
                                repeat=True,
                                snr=UniversalSNR(value=snr.value, is_random=snr.is_random),
                            ),
                        },
                        samples=primary_length,
                        spectral_mask_id=spectral_mask_id + 1,
                        spectral_mask_seed=randint(0, np.iinfo("i").max),  # noqa: S311
                    )
                )
                m_id += 1

    return mixtures


def _non_combinatorial_noise_mix(
    location: str,
    config: dict,
    effected_sources: dict[str, list[tuple[SourceFile, Effects]]],
    test: bool = False,
) -> list[Mixture]:
    """Combine a source/effect+interferences/effect with a single cut of a noise/effect
    non-exhaustively (each source/effect+interferences/effect does not use each noise/effect).
    Cut has random start and loop back to beginning if end of noise/effect is reached.
    """
    from random import choice
    from random import randint

    import numpy as np

    from ..datatypes import Mixture
    from ..datatypes import UniversalSNR
    from .effects import effects_from_rules
    from .effects import estimate_effected_length

    mixdb = MixtureDatabase(location, test)
    snrs = get_all_snrs_from_config(config)

    m_id = 0
    noise_id = 0
    mixtures: list[Mixture] = []
    for primary_file, primary_rule in effected_sources["primary"]:
        primary_effect = effects_from_rules(mixdb, primary_rule)
        primary_length = estimate_effected_length(primary_file.samples, primary_effect, mixdb.feature_step_samples)

        for spectral_mask_id in range(len(config["spectral_masks"])):
            for snr in snrs["noise"]:
                noise_file, noise_rule = effected_sources["noise"][noise_id]
                noise_effect = effects_from_rules(mixdb, noise_rule)
                noise_length = estimate_effected_length(noise_file.samples, noise_effect)

                mixtures.append(
                    Mixture(
                        name=str(m_id),
                        all_sources={
                            "primary": Source(
                                file_id=primary_file.id,
                                effects=primary_effect,
                            ),
                            "noise": Source(
                                file_id=noise_file.id,
                                effects=noise_effect,
                                start=choice(range(noise_length)),  # noqa: S311
                                repeat=True,
                                snr=UniversalSNR(value=snr.value, is_random=snr.is_random),
                            ),
                        },
                        samples=primary_length,
                        spectral_mask_id=spectral_mask_id + 1,
                        spectral_mask_seed=randint(0, np.iinfo("i").max),  # noqa: S311
                    )
                )
                noise_id = (noise_id + 1) % len(effected_sources["noise"])
                m_id += 1

    return mixtures


class NextNoise:
    def __init__(self, mixdb: MixtureDatabase, effected_noises: list[tuple[SourceFile, Effects]]) -> None:
        from .effects import effects_from_rules
        from .effects import estimate_effected_length

        self.mixdb = mixdb
        self.effected_noises = effected_noises

        self.noise_start = 0
        self.noise_id = 0
        self.noise_effect = effects_from_rules(self.mixdb, self.noise_rule)
        self.noise_length = estimate_effected_length(self.noise_file.samples, self.noise_effect)

    @property
    def noise_file(self):
        return self.effected_noises[self.noise_id][0]

    @property
    def noise_rule(self):
        return self.effected_noises[self.noise_id][1]

    def generate(self, length: int) -> tuple[int, Effects, int]:
        from .effects import effects_from_rules
        from .effects import estimate_effected_length

        if self.noise_start + length > self.noise_length:
            # Not enough samples in current noise
            if self.noise_start == 0:
                raise ValueError("Length of primary audio exceeds length of noise audio")

            self.noise_start = 0
            self.noise_id = (self.noise_id + 1) % len(self.effected_noises)
            self.noise_effect = effects_from_rules(self.mixdb, self.noise_rule)
            self.noise_length = estimate_effected_length(self.noise_file.samples, self.noise_effect)
            noise_start = self.noise_start
        else:
            # Current noise has enough samples
            noise_start = self.noise_start
            self.noise_start += length

        return self.noise_file.id, self.noise_effect, noise_start


def get_all_snrs_from_config(config: dict) -> dict[str, list[UniversalSNRGenerator]]:
    snrs: dict[str, list[UniversalSNRGenerator]] = {}
    for category in config["sources"]:
        if category != "primary":
            snrs[category] = [UniversalSNRGenerator(snr) for snr in config["sources"][category]["snrs"]]
    return snrs


def _get_textgrid_tiers_from_source_file(file: str) -> list[str]:
    from pathlib import Path

    from praatio import textgrid

    from ..utils.tokenized_shell_vars import tokenized_expand

    textgrid_file = Path(tokenized_expand(file)[0]).with_suffix(".TextGrid")
    if not textgrid_file.exists():
        return []

    tg = textgrid.openTextgrid(str(textgrid_file), includeEmptyIntervals=False)

    return sorted(tg.tierNames)


def _populate_speaker_table(
    location: str,
    source_files: list[SourceFile],
    test: bool = False,
    verbose: bool = False,
) -> None:
    """Populate speaker table"""
    import json
    from pathlib import Path

    import yaml

    from ..utils.tokenized_shell_vars import tokenized_expand
    from .mixdb import db_connection

    # Determine columns for speaker table
    all_parents = {Path(file.name).parent for file in source_files}
    speaker_parents = (parent for parent in all_parents if Path(tokenized_expand(parent / "speaker.yml")[0]).exists())

    speakers: dict[Path, dict[str, str]] = {}
    for parent in sorted(speaker_parents):
        with open(tokenized_expand(parent / "speaker.yml")[0]) as f:
            speakers[parent] = yaml.safe_load(f)

    new_columns: list[str] = []
    for keys in speakers:
        for column in speakers[keys]:
            new_columns.append(column)
    new_columns = sorted(set(new_columns))

    con = db_connection(location=location, readonly=False, test=test, verbose=verbose)

    for new_column in new_columns:
        con.execute(f"ALTER TABLE speaker ADD COLUMN {new_column} TEXT")

    # Populate speaker table
    speaker_rows: list[tuple[str, ...]] = []
    for key in speakers:
        entry = (speakers[key].get(column, None) for column in new_columns)
        speaker_rows.append((key.as_posix(), *entry))  # type: ignore[arg-type]

    column_ids = ", ".join(["parent", *new_columns])
    column_values = ", ".join(["?"] * (len(new_columns) + 1))
    con.executemany(f"INSERT INTO speaker ({column_ids}) VALUES ({column_values})", speaker_rows)

    con.execute("CREATE INDEX speaker_parent_idx ON speaker (parent)")

    # Update speaker_metadata_tiers in the top table
    tiers = [
        description[0]
        for description in con.execute("SELECT * FROM speaker").description
        if description[0] not in ("id", "parent")
    ]
    con.execute("UPDATE top SET speaker_metadata_tiers=? WHERE ? = id", (json.dumps(tiers), 1))

    if "speaker_id" in tiers:
        con.execute("CREATE INDEX speaker_speaker_id_idx ON source_file (speaker_id)")

    con.commit()
    con.close()


def _populate_truth_config_table(
    location: str,
    source_files: list[SourceFile],
    test: bool = False,
    verbose: bool = False,
) -> None:
    """Populate truth_config table"""
    import json

    from .mixdb import db_connection

    con = db_connection(location=location, readonly=False, test=test, verbose=verbose)

    # Populate truth_config table
    truth_configs: list[str] = []
    for file in source_files:
        for name, config in file.truth_configs.items():
            ts = json.dumps({"name": name} | config.to_dict())
            if ts not in truth_configs:
                truth_configs.append(ts)
    con.executemany(
        "INSERT INTO truth_config (config) VALUES (?)",
        [(item,) for item in truth_configs],
    )

    con.commit()
    con.close()


def _populate_impulse_response_tag_table(
    location: str,
    files: list[ImpulseResponseFile],
    test: bool = False,
    verbose: bool = False,
) -> None:
    """Populate ir_tag table"""
    from .mixdb import db_connection

    con = db_connection(location=location, readonly=False, test=test, verbose=verbose)

    con.executemany(
        "INSERT INTO ir_tag (tag) VALUES (?)",
        [(tag,) for tag in {tag for file in files for tag in file.tags}],
    )

    con.commit()
    con.close()



from dataclasses import dataclass


@dataclass(frozen=True)
class TimeAlignedType:
    start: float
    end: float
    text: str

    @property
    def duration(self) -> float:
        return self.end - self.start



from ...datatypes import Truth
from ..mixdb import MixtureDatabase


def phoneme_validate(_config: dict) -> None:
    raise NotImplementedError("Truth function phoneme is not supported yet")


def phoneme_parameters(_feature: str, _num_classes: int, _config: dict) -> int:
    raise NotImplementedError("Truth function phoneme is not supported yet")


def phoneme(_mixdb: MixtureDatabase, _m_id: int, _category: str, _config: dict) -> Truth:
    """Read in .txt transcript and run a Python function to generate text grid data
    (indicating which phonemes are active). Then generate truth based on this data and put
    in the correct classes based on the index in the config.
    """
    raise NotImplementedError("Truth function phoneme is not supported yet")



"""sonusai metrics_summary

usage: lsdb [-vlh] [-i MIXID] [-n NCPU] LOCATION

Options:
    -h, --help
    -v, --verbose
    -l, --write-list            Write .csv file list of all mixture metrics
    -i MIXID, --mixid MIXID     Mixture ID(s) to analyze. [default: *].
    -n, --num_process NCPU      Number of parallel processes to use [default: auto]

Summarize mixture metrics across a SonusAI mixture database where metrics have been generated by SonusAI genmetrics.

Inputs:
    LOCATION     A SonusAI mixture database directory with mixdb.db and pre-generated metrics from SonusAI genmetrics.

"""

import numpy as np
import pandas as pd

DB_99 = np.power(10, 99 / 10)
DB_N99 = np.power(10, -99 / 10)


def _process_mixture(
    m_id: int,
    location: str,
    all_metric_names: list[str],
    scalar_metric_names: list[str],
    string_metric_names: list[str],
    frame_metric_names: list[str],
    bin_metric_names: list[str],
    ptab_labels: list[str],
) -> tuple[pd.DataFrame, pd.DataFrame]:
    from os.path import basename

    from sonusai.constants import SAMPLE_RATE
    from sonusai.metrics import calc_wer
    from sonusai.mixture import MixtureDatabase

    mixdb = MixtureDatabase(location)

    # Process mixture
    # for mixid in mixids:
    samples = mixdb.mixture(m_id).samples
    duration = samples / SAMPLE_RATE
    tf_frames = mixdb.mixture_transform_frames(m_id)
    feat_frames = mixdb.mixture_feature_frames(m_id)
    mxsnr = mixdb.mixture(m_id).noise.snr
    ti = mixdb.mixture(m_id).sources["primary"].file_id
    ni = mixdb.mixture(m_id).noise.file_id
    t0file = basename(mixdb.source_file(ti).name)
    nfile = basename(mixdb.source_file(ni).name)

    all_metrics = mixdb.mixture_metrics(m_id, all_metric_names)

    # replace lists with first value (ignore mixup)
    scalar_metrics = {
        key: all_metrics[key][0] if isinstance(all_metrics[key], list) else all_metrics[key]
        for key in scalar_metric_names
    }
    string_metrics = {
        key: all_metrics[key][0] if isinstance(all_metrics[key], list) else all_metrics[key]
        for key in string_metric_names
    }

    # Convert strings into word count
    for key in string_metrics:
        string_metrics[key] = calc_wer(string_metrics[key], string_metrics[key]).words

    # Collect pandas table values note: must match given ptab_labels
    ptab_data: list = [
        mxsnr,
        *scalar_metrics.values(),
        *string_metrics.values(),
        tf_frames,
        duration,
        t0file,
        nfile,
    ]

    ptab1 = pd.DataFrame([ptab_data], columns=ptab_labels, index=[m_id])

    # TODO: collect frame metrics and bin metrics

    return ptab1, ptab1


def main() -> None:
    from docopt import docopt

    from . import __version__ as sai_version
    from .utils.docstring import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sai_version, options_first=True)

    verbose = args["--verbose"]
    wrlist = args["--write-list"]
    mixids = args["--mixid"]
    location = args["LOCATION"]
    num_proc = args["--num_process"]

    from functools import partial
    from os.path import basename
    from os.path import join

    import psutil

    from . import create_file_handler
    from . import initial_log_messages
    from . import logger
    from . import update_console_handler
    from .mixture.mixdb import MixtureDatabase
    from .utils.create_timestamp import create_timestamp
    from .utils.parallel import par_track
    from .utils.parallel import track

    mixdb = MixtureDatabase(location)
    print(f"Found SonusAI mixture database with {mixdb.num_mixtures} mixtures.")

    # Only check first and last mixture in order to save time
    metrics_present = mixdb.cached_metrics([0, mixdb.num_mixtures - 1])  # return pre-generated metrics in mixdb tree
    if "mxsnr" in metrics_present:
        metrics_present.remove("mxsnr")

    num_metrics_present = len(metrics_present)
    if num_metrics_present < 1:
        print(f"mixdb reports no pre-generated metrics are present. Nothing to summarize in {location}, exiting ...")
        return

    # Setup logging file
    timestamp = create_timestamp()  # string good for embedding into filenames
    mixdb_fname = basename(location)
    if verbose:
        create_file_handler(join(location, "metrics_summary.log"))
        update_console_handler(verbose)
        initial_log_messages("metrics_summary")
        logger.info(f"Logging summary of SonusAI mixture database at {location}")
    else:
        update_console_handler(verbose)

    logger.info("")
    mixids = mixdb.mixids_to_list(mixids)
    if len(mixids) < mixdb.num_mixtures:
        logger.info(
            f"Processing a subset of {len(mixids)} out of total mixdb mixtures of {mixdb.num_mixtures}, "
            f"summary results will not include entire dataset."
        )
        fsuffix = f"_s{len(mixids)}t{mixdb.num_mixtures}"
    else:
        logger.info(
            f"Summarizing SonusAI mixture database with {mixdb.num_mixtures} mixtures "
            f"and {num_metrics_present} pre-generated metrics ..."
        )
        fsuffix = ""

    metric_sup = mixdb.supported_metrics
    ft_bins = mixdb.ft_config.bin_end - mixdb.ft_config.bin_start + 1  # bins of forward transform
    # Pre-process first mixid to gather metrics into 4 types: scalar, str (scalar word cnt), frame-array, bin-array
    # Collect list of indices for each
    scalar_metric_names: list[str] = []
    string_metric_names: list[str] = []
    frame_metric_names: list[str] = []
    bin_metric_names: list[str] = []
    all_metrics = mixdb.mixture_metrics(mixids[0], metrics_present)
    tf_frames = mixdb.mixture_transform_frames(mixids[0])
    for metric in metrics_present:
        metval = all_metrics[metric]  # get metric value
        logger.debug(f"First mixid {mixids[0]} metric {metric} = {metval}")
        if isinstance(metval, list):
            if len(metval) > 1:
                logger.warning(f"Mixid {mixids[0]} metric {metric} has a list with more than 1 element, using first.")
            metval = metval[0]  # remove any list
        if isinstance(metval, float | int):
            logger.debug(f"Metric is scalar {type(metval)}, entering in summary table.")
            scalar_metric_names.append(metric)
        elif isinstance(metval, str):
            logger.debug("Metric is string, will summarize with word count.")
            string_metric_names.append(metric)
        elif isinstance(metval, np.ndarray):
            if metval.ndim == 1:
                if metval.size == tf_frames:
                    logger.debug("Metric is frames vector.")
                    frame_metric_names.append(metric)
                elif metval.size == ft_bins:
                    logger.debug("Metric is bins vector.")
                    bin_metric_names.append(metric)
                else:
                    logger.warning(f"Mixid {mixids[0]} metric {metric} is a vector of improper size, ignoring.")

    # Setup pandas table for summarizing scalar metrics, always include mxsnr first
    ptab_labels = [
        "mxsnr",
        *scalar_metric_names,
        *string_metric_names,
        "fcnt",
        "duration",
        "t0file",
        "nfile",
    ]

    num_cpu = psutil.cpu_count()
    cpu_percent = psutil.cpu_percent(interval=1)
    logger.info("")
    logger.info(f"#CPUs: {num_cpu}, current CPU utilization: {cpu_percent}%")
    logger.info(f"Memory utilization: {psutil.virtual_memory().percent}%")
    if num_proc == "auto":
        use_cpu = int(num_cpu * (0.9 - cpu_percent / 100))  # default use 80% of available cpus
    elif num_proc == "None":
        use_cpu = None
    else:
        use_cpu = min(max(int(num_proc), 1), num_cpu)

    logger.info(f"Summarizing metrics for {len(mixids)} mixtures using {use_cpu} parallel processes")

    # progress = tqdm(total=len(mixids), desc='calc_metric_spenh', mininterval=1)
    progress = track(total=len(mixids))
    if use_cpu is None:
        no_par = True
        num_cpus = None
    else:
        no_par = False
        num_cpus = use_cpu

    all_metrics_tables = par_track(
        partial(
            _process_mixture,
            location=location,
            all_metric_names=metrics_present,
            scalar_metric_names=scalar_metric_names,
            string_metric_names=string_metric_names,
            frame_metric_names=frame_metric_names,
            bin_metric_names=bin_metric_names,
            ptab_labels=ptab_labels,
        ),
        mixids,
        progress=progress,
        num_cpus=num_cpus,
        no_par=no_par,
    )
    progress.close()

    # Done with mixtures, write out summary metrics
    header_args = {
        "mode": "a",
        "encoding": "utf-8",
        "index": False,
        "header": False,
    }
    table_args = {
        "mode": "a",
        "encoding": "utf-8",
    }
    ptab1 = pd.concat([item[0] for item in all_metrics_tables])
    if wrlist:
        wlcsv_name = str(join(location, "metric_summary_list" + fsuffix + ".csv"))
        pd.DataFrame([["Timestamp", timestamp]]).to_csv(wlcsv_name, header=False, index=False)
        pd.DataFrame([f"Metric list for {mixdb_fname}:"]).to_csv(wlcsv_name, mode="a", header=False, index=False)
        ptab1.round(2).to_csv(wlcsv_name, **table_args)
    ptab1_sorted = ptab1.sort_values(by=["mxsnr", "t0file"])

    # Create metrics table except -99 SNR
    ptab1_nom99 = ptab1_sorted[ptab1_sorted.mxsnr != -99]

    # Create summary by SNR for all scalar metrics, taking mean
    mtab_snr_summary = None
    for snri in range(0, len(mixdb.snrs)):
        tmp = ptab1_sorted.query("mxsnr==" + str(mixdb.snrs[snri])).mean(numeric_only=True).to_frame().T
        # avoid nan when subset of mixids specified (i.e. no mixtures exist for an SNR)
        if ~np.isnan(tmp.iloc[0].to_numpy()[0]).any():
            mtab_snr_summary = pd.concat([mtab_snr_summary, tmp])
    mtab_snr_summary = mtab_snr_summary.sort_values(by=["mxsnr"], ascending=False)

    # Write summary to .csv
    snrcsv_name = str(join(location, "metric_summary_snr" + fsuffix + ".csv"))
    nmix = len(mixids)
    nmixtot = mixdb.num_mixtures
    pd.DataFrame([["Timestamp", timestamp]]).to_csv(snrcsv_name, header=False, index=False)
    pd.DataFrame(['"Metrics avg over each SNR:"']).to_csv(snrcsv_name, **header_args)
    mtab_snr_summary.round(2).T.to_csv(snrcsv_name, index=True, header=False, mode="a", encoding="utf-8")
    pd.DataFrame(["--"]).to_csv(snrcsv_name, header=False, index=False, mode="a")
    pd.DataFrame([f'"Metrics stats over {nmix} mixtures out of {nmixtot} total:"']).to_csv(snrcsv_name, **header_args)
    ptab1.describe().round(2).T.to_csv(snrcsv_name, index=True, **table_args)
    pd.DataFrame(["--"]).to_csv(snrcsv_name, header=False, index=False, mode="a")
    pd.DataFrame([f'"Metrics stats over {len(ptab1_nom99)} non -99db mixtures out of {nmixtot} total:"']).to_csv(
        snrcsv_name, **header_args
    )
    ptab1_nom99.describe().round(2).T.to_csv(snrcsv_name, index=True, **table_args)

    # Write summary to text file
    snrtxt_name = str(join(location, "metric_summary_snr" + fsuffix + ".txt"))
    with open(snrtxt_name, "w") as f:
        print(f"Timestamp: {timestamp}", file=f)
        print("Metrics avg over each SNR:", file=f)
        print(
            mtab_snr_summary.round(2).T.to_string(float_format=lambda x: f"{x:.2f}", index=True, header=False), file=f
        )
        print("", file=f)
        print(f"Metrics stats over {len(mixids)} mixtures out of {mixdb.num_mixtures} total:", file=f)
        print(ptab1.describe().round(2).T.to_string(float_format=lambda x: f"{x:.2f}", index=True), file=f)
        print("", file=f)
        print(f"Metrics stats over {len(ptab1_nom99)} non -99db mixtures out of {mixdb.num_mixtures} total:", file=f)
        print(ptab1_nom99.describe().round(2).T.to_string(float_format=lambda x: f"{x:.2f}", index=True), file=f)


if __name__ == "__main__":
    from sonusai import exception_handler
    from sonusai.utils import register_keyboard_interrupt

    register_keyboard_interrupt()
    try:
        main()
    except Exception as e:
        exception_handler(e)
